#
# This file is part of `gitflow`.
# Copyright (c) 2010-2011 Vincent Driessen
# Copyright (c) 2012 Hartmut Goebel
# Distributed under a BSD-like license. For full terms see the file LICENSE.txt
#

import sys
import re
from functools import wraps
try:
    from cStringIO import StringIO
except:
    from StringIO import StringIO

from unittest2 import TestCase

import gitflow
from gitflow.core import GitFlow, Repo
from gitflow.exceptions import (NoSuchBranchError, NoSuchRemoteError,
                                AlreadyInitialized, NotInitialized,
                                BaseNotOnBranch)
from gitflow.bin import (main as Main,
                         VersionCommand, StatusCommand, InitCommand,
                         FeatureCommand, ReleaseCommand, HotfixCommand,
                         SupportCommand
                         )

from gitflow.branches import BranchManager
from tests.helpers import (copy_from_fixture, remote_clone_from_fixture,
                           all_commits, sandboxed, fake_commit)
from tests.helpers.factory import create_sandbox, create_git_repo

__copyright__ = "2010-2011 Vincent Driessen; 2012 Hartmut Goebel"
__license__ = "BSD"

def runGitFlow(*argv, **kwargs):
    capture = kwargs.get('capture', False)
    _argv, sys.argv = sys.argv, ['git-flow'] + list(argv)
    _stdout = sys.stdout
    try:
        if not capture:
            gitflow.bin.main()
        else:
            sys.stdout = StringIO()
            gitflow.bin.main()
            return sys.stdout.getvalue()
    finally:
        sys.stdout = _stdout
        sys.argv = _argv


class TestCase(TestCase):
    def assertArgparseError(self, expected_regexp, func, *args, **kwargs):
        _stderr, sys.stderr = sys.stderr, StringIO()
        try:
            self.assertRaises(SystemExit, func, *args, **kwargs)
            msg = sys.stderr.getvalue()
            expected_regexp = re.compile(expected_regexp)
            if not expected_regexp.search(str(msg)):
                raise self.failureException('"%s" does not match "%s"' %
                         (expected_regexp.pattern, msg))
        finally:
            sys.stderr = _stderr

class TestVersionCommand(TestCase):

    def test_version(self):
        stdout = runGitFlow('version', capture=1)
        self.assertEqual(gitflow.__version__+'\n', stdout)


class TestStatusCommand(TestCase):

    @copy_from_fixture('sample_repo')
    def test_version(self):
        stdout = runGitFlow('status', capture=1)
        self.assertItemsEqual([
            '  devel: 2b34cd2',
            '  feat/even: e56be18',
            '* feat/recursion: 54d59c8',
            '  stable: 296586b',
            ], stdout.splitlines())


class TestInitCommand(TestCase):
    @sandboxed
    def test_init_defaults(self):
        runGitFlow('init', '--defaults')
        gitflow = GitFlow('.').init()
        self.assertEquals('origin', gitflow.origin_name())
        self.assertEquals('master', gitflow.master_name())
        self.assertEquals('develop', gitflow.develop_name())
        self.assertEquals('feature/', gitflow.get_prefix('feature'))
        self.assertEquals('release/', gitflow.get_prefix('release'))
        self.assertEquals('hotfix/', gitflow.get_prefix('hotfix'))
        self.assertEquals('support/', gitflow.get_prefix('support'))
        self.assertEquals('', gitflow.get_prefix('versiontag'))

    @sandboxed
    def test_init_accepting_defaults(self):
        text = '\n'*8
        _stdin, sys.stdin = sys.stdin, StringIO(text)
        try:
            runGitFlow('init')
        finally:
            sys.stdin = _stdin
        gitflow = GitFlow('.').init()
        self.assertEquals('origin', gitflow.origin_name())
        self.assertEquals('master', gitflow.master_name())
        self.assertEquals('develop', gitflow.develop_name())
        self.assertEquals('feature/', gitflow.get_prefix('feature'))
        self.assertEquals('release/', gitflow.get_prefix('release'))
        self.assertEquals('hotfix/', gitflow.get_prefix('hotfix'))
        self.assertEquals('support/', gitflow.get_prefix('support'))
        self.assertEquals('', gitflow.get_prefix('versiontag'))

    @sandboxed
    def test_init_custom(self):
        text = '\n'.join(['my-remote', 'stable', 'devel',
                          'feat/', 'rel/', 'hf/', 'sup/', 'ver'])
        _stdin, sys.stdin = sys.stdin, StringIO(text)
        try:
            runGitFlow('init')
        finally:
            sys.stdin = _stdin
        gitflow = GitFlow('.').init()
        self.assertEquals('my-remote', gitflow.origin_name())
        self.assertEquals('stable', gitflow.master_name())
        self.assertEquals('devel', gitflow.develop_name())
        self.assertEquals('feat/', gitflow.get_prefix('feature'))
        self.assertEquals('rel/', gitflow.get_prefix('release'))
        self.assertEquals('hf/', gitflow.get_prefix('hotfix'))
        self.assertEquals('sup/', gitflow.get_prefix('support'))
        self.assertEquals('ver', gitflow.get_prefix('versiontag'))

    @sandboxed
    def test_init_custom_accepting_some_defaults(self):
        text = '\n'.join(['my-remote', '', 'devel',
                          'feat/', '', '', 'sup/', 'v'])
        _stdin, sys.stdin = sys.stdin, StringIO(text)
        try:
            runGitFlow('init')
        finally:
            sys.stdin = _stdin
        gitflow = GitFlow('.').init()
        self.assertEquals('my-remote', gitflow.origin_name())
        self.assertEquals('master', gitflow.master_name())
        self.assertEquals('devel', gitflow.develop_name())
        self.assertEquals('feat/', gitflow.get_prefix('feature'))
        self.assertEquals('release/', gitflow.get_prefix('release'))
        self.assertEquals('hotfix/', gitflow.get_prefix('hotfix'))
        self.assertEquals('sup/', gitflow.get_prefix('support'))
        self.assertEquals('v', gitflow.get_prefix('versiontag'))


    @copy_from_fixture('custom_repo')
    def test_init_fails_if_already_initialized(self):
        self.assertRaises(AlreadyInitialized, runGitFlow, 'init')


    @copy_from_fixture('custom_repo')
    def test_init_force_defaults_succeeds_if_already_initialized(self):
        runGitFlow('init', '--defaults', '--force')
        gitflow = GitFlow('.').init()
        # these are the values already defined in custom_repo
        self.assertEquals('origin', gitflow.origin_name())
        self.assertEquals('production', gitflow.master_name())
        self.assertEquals('master', gitflow.develop_name())
        self.assertEquals('f-', gitflow.get_prefix('feature'))
        self.assertEquals('rel-', gitflow.get_prefix('release'))
        self.assertEquals('hf-', gitflow.get_prefix('hotfix'))
        self.assertEquals('supp-', gitflow.get_prefix('support'))
        self.assertEquals('v', gitflow.get_prefix('versiontag'))

    @copy_from_fixture('custom_repo')
    def test_init_existing_repo_fails_on_non_existing_master_branch(self):
        text = '\n'.join(['', 'stable', '', '', '', '', '', ''])
        _stdin, sys.stdin = sys.stdin, StringIO(text)
        try:
            self.assertRaises(NoSuchBranchError, runGitFlow, 'init', '--force')
        finally:
            sys.stdin = _stdin

    @copy_from_fixture('custom_repo')
    def test_init_existing_repo_fails_on_non_existing_develop_branch(self):
        text = '\n'.join(['', '', 'workinprogress', '', '', '', '', ''])
        _stdin, sys.stdin = sys.stdin, StringIO(text)
        try:
            self.assertRaises(NoSuchBranchError, runGitFlow, 'init', '--force')
        finally:
            sys.stdin = _stdin

    @copy_from_fixture('custom_repo')
    def test_init_force_succeeds_if_already_initialized(self):
        # NB: switching master and develop
        text = '\n'.join(['my-remote', 'master', 'production',
                          'feat/', 'rel/', 'hf/', 'sup/', 'ver'])
        _stdin, sys.stdin = sys.stdin, StringIO(text)
        try:
            runGitFlow('init', '--force')
        finally:
            sys.stdin = _stdin
        gitflow = GitFlow('.').init()
        self.assertEquals('master', gitflow.master_name())
        self.assertEquals('production', gitflow.develop_name())

    @sandboxed
    def test_init_fails_if_develop_name_equals_master_name(self):
        text = '\n'.join(['', 'mainbranch', 'mainbranch'])
        _stdin, sys.stdin = sys.stdin, StringIO(text)
        try:
            self.assertRaisesRegexp(SystemExit, ".*branches should differ.*",
                                    runGitFlow, 'init')
        finally:
            sys.stdin = _stdin


    @remote_clone_from_fixture('sample_repo')
    def test_init_with_master_existing_remote_but_not_local(self):
        rsc0 = self.remote.branches['stable'].commit
        text = '\n'.join(['my-remote', 'stable', 'devel',
                          'feat/', 'rel/', 'hf/', 'sup/', 'ver'])
        _stdin, sys.stdin = sys.stdin, StringIO(text)
        try:
            runGitFlow('init', '--force')
        finally:
            sys.stdin = _stdin
        rsc1 = self.remote.branches['stable'].commit
        lsc1 = self.repo.branches['stable'].commit
        self.assertEqual(rsc0, rsc1)
        self.assertEqual(rsc1, lsc1)


    # These tests need a repo with only branches `foo` and `bar`
    # or other names not selected for defaults
    # :todo: give no master branch name (or white-spaces)
    # :todo: give no develop branch name (or white-spaces)


    @remote_clone_from_fixture('sample_repo')
    def test_subcommands_requiring_initialized_repo(self):
        def assertSystemExit(*args):
            self.assertRaisesRegexp(
                SystemExit, 'repo has not yet been initialized for git-flow',
                runGitFlow, 'feature', *args)
        def assertNotInitialized(*args):
            self.assertRaises(NotInitialized,
                runGitFlow, 'feature', *args)
        assertNotInitialized('start', 'xxx')
        assertNotInitialized('finish', 'recursion')
        assertNotInitialized('publish', 'recursion')
        assertNotInitialized('track', 'recursion')
        assertNotInitialized('diff', 'recursion')
        assertNotInitialized('rebase', 'recursion')
        assertNotInitialized('pull', 'even')

class TestFeature(TestCase):

    @copy_from_fixture('sample_repo')
    def test_feature_list(self):
        stdout = runGitFlow('feature', 'list', capture=1)
        expected = [
          '  even',
          '* recursion'
          ]
        self.assertEqual(stdout.splitlines(), expected)

    @copy_from_fixture('sample_repo')
    def test_feature_list_verbose(self):
        stdout = runGitFlow('feature', 'list', '--verbose', capture=1)
        expected = [
          '  even      (based on latest devel)',
          '* recursion (based on latest devel)'
          ]
        self.assertEqual(stdout.splitlines(), expected)

    @copy_from_fixture('sample_repo')
    def test_feature_list_verbose_rebased(self):
        self.repo.refs['devel'].checkout()
        fake_commit(self.repo, 'A commit on devel')
        stdout = runGitFlow('feature', 'list', '--verbose', capture=1)
        expected = [
          '  even      (may be rebased)',
          '  recursion (may be rebased)'
          ]
        self.assertEqual(stdout.splitlines(), expected)

    @sandboxed
    def test_feature_list_verbose_no_commits(self):
        repo = create_git_repo(self)
        gitflow = GitFlow(repo).init()
        repo.create_head('feature/wow', 'HEAD')
        stdout = runGitFlow('feature', 'list', '--verbose', capture=1)
        expected = [
          '  wow (no commits yet)'
          ]
        self.assertEqual(stdout.splitlines(), expected)

    @copy_from_fixture('sample_repo')
    def test_feature_list_verbose_ff(self):
        self.repo.create_head('devel', 'feat/recursion', force=1)
        self.repo.refs['devel'].checkout()
        fake_commit(self.repo, 'A commit on devel')
        stdout = runGitFlow('feature', 'list', '--verbose', capture=1)
        expected = [
          '  even      (may be rebased)',
          '  recursion (is behind devel, may ff)'
          ]
        self.assertEqual(stdout.splitlines(), expected)

    #--- feature start ---

    @copy_from_fixture('sample_repo')
    def test_feature_start(self):
        runGitFlow('feature', 'start', 'wow-feature')
        self.assertIn('feat/wow-feature', Repo().branches)

    @copy_from_fixture('sample_repo')
    def test_feature_start_alt_base(self):
        runGitFlow('feature', 'start', 'wow-feature', 'devel')
        self.assertIn('feat/wow-feature', Repo().branches)

    ## feature branch is not required to start at `develop`, rethink this
    ## @copy_from_fixture('sample_repo')
    ## def test_feature_start_wrong_alt_base_raises_error(self):
    ##     self.repo.refs['stable'].checkout()
    ##     fake_commit(self.repo, 'A fake commit on stable')
    ##     self.assertRaises(BaseNotOnBranch,
    ##                       runGitFlow, 'feature', 'start', 'wow', 'stable')

    @copy_from_fixture('sample_repo')
    def test_feature_start_empty_name(self):
        self.assertArgparseError('must not by empty',
                                 runGitFlow, 'feature', 'start', '')
        self.assertArgparseError('must not by empty',
                                 runGitFlow, 'feature', 'start', '', 'devel')

    @copy_from_fixture('sample_repo')
    def test_feature_start_no_name(self):
        self.assertArgparseError('too few arguments',
                                 runGitFlow, 'feature', 'start')

    #--- feature finish ---

    @remote_clone_from_fixture('sample_repo')
    def test_feature_finish(self):
        GitFlow('.').init()
        runGitFlow('feature', 'finish', 'recursion')
        self.assertNotIn('feat/recursion', Repo().branches)

    @copy_from_fixture('sample_repo')
    def test_feature_finish_current(self):
        GitFlow('.').init()
        runGitFlow('feature', 'finish')
        self.assertNotIn('feat/recursion', Repo().branches)

    @copy_from_fixture('sample_repo')
    def test_feature_finish_empty_prefix(self):
        GitFlow('.').init()
        runGitFlow('feature', 'checkout', 'even')
        runGitFlow('feature', 'finish', '')
        self.assertNotIn('feat/even', Repo().branches)

    @copy_from_fixture('sample_repo')
    def test_feature_finish_prefix(self):
        GitFlow('.').init()
        runGitFlow('feature', 'finish', 'e')
        self.assertNotIn('feat/even', Repo().branches)

    @copy_from_fixture('sample_repo')
    def test_feature_finish_rebase(self):
        gitflow = GitFlow('.').init()
        gitflow.develop().checkout()
        fake_commit(gitflow.repo, 'A commit on devel')
        runGitFlow('feature', 'finish', 'even', '--rebase')
        self.assertNotIn('feat/even', Repo().branches)
        self.assertEqual(gitflow.develop().commit.message,
                         'Finished feature even.\n')
        # :todo: think about some other test to see if it really worked

    #:todo: test-cases for `feature finish --rebase` w/ conflict

    @copy_from_fixture('sample_repo')
    def test_feature_finish_keep(self):
        GitFlow('.').init()
        runGitFlow('feature', 'finish', 'even', '--keep')
        self.assertIn('feat/even', Repo().branches)

    #:todo: test-cases for `feature finish --fetch`
    #:todo: test-cases for `feature finish --force-delete`

    #--- feature publish ---

    @remote_clone_from_fixture('sample_repo')
    def test_feature_publish(self):
        GitFlow('.').init()
        runGitFlow('feature', 'start', 'wow')
        runGitFlow('feature', 'publish', 'wow')
        self.assertIn('feat/wow', Repo().remotes['my-remote'].refs)

    @remote_clone_from_fixture('sample_repo')
    def test_feature_publish_current(self):
        GitFlow('.').init()
        runGitFlow('feature', 'start', 'wow')
        runGitFlow('feature', 'publish')
        self.assertIn('feat/wow', Repo().remotes['my-remote'].refs)

    @remote_clone_from_fixture('sample_repo')
    def test_feature_publish_empty_prefix(self):
        GitFlow('.').init()
        runGitFlow('feature', 'start', 'wow')
        runGitFlow('feature', 'publish', '')
        self.assertIn('feat/wow', Repo().remotes['my-remote'].refs)

    @remote_clone_from_fixture('sample_repo')
    def test_feature_publish_prefix(self):
        GitFlow('.').init()
        runGitFlow('feature', 'start', 'wow')
        runGitFlow('feature', 'publish', 'w')
        self.assertIn('feat/wow', Repo().remotes['my-remote'].refs)

    #--- feature track ---

    @remote_clone_from_fixture('sample_repo')
    def test_feature_track(self):
        GitFlow('.').init()
        runGitFlow('feature', 'track', 'even')

    @remote_clone_from_fixture('sample_repo')
    def test_feature_track_name_is_required(self):
        GitFlow('.').init()
        self.assertArgparseError('too few arguments',
                                 runGitFlow, 'feature', 'track')
        self.assertArgparseError('must not by empty',
                                 runGitFlow, 'feature', 'track', '')

    #--- feature diff ---

    @copy_from_fixture('sample_repo')
    def test_feature_diff(self):
        runGitFlow('feature', 'diff', 'recursion')

    @copy_from_fixture('sample_repo')
    def test_feature_diff_current(self):
        runGitFlow('feature', 'diff')

    @copy_from_fixture('sample_repo')
    def test_feature_diff_empty_prefix(self):
        runGitFlow('feature', 'diff', '')

    @copy_from_fixture('sample_repo')
    def test_feature_diff_prefix(self):
        runGitFlow('feature', 'diff', 'rec')

    #--- feature rebase ---

    @copy_from_fixture('sample_repo')
    def test_feature_rebase(self):
        runGitFlow('feature', 'rebase', 'recursion')

    @copy_from_fixture('sample_repo')
    def test_feature_rebase_current(self):
        runGitFlow('feature', 'rebase')

    @copy_from_fixture('sample_repo')
    def test_feature_rebase_empty_prefix(self):
        runGitFlow('feature', 'rebase', '')

    @copy_from_fixture('sample_repo')
    def test_feature_rebase_prefix(self):
        runGitFlow('feature', 'rebase', 'rec')

    #--- feature  checkout ---

    @copy_from_fixture('sample_repo')
    def test_feature_checkout(self):
        runGitFlow('feature', 'checkout', 'even')

    @copy_from_fixture('sample_repo')
    def test_feature_checkout_current(self):
        self.assertArgparseError('too few arguments',
                                 runGitFlow, 'feature', 'checkout')

    @copy_from_fixture('sample_repo')
    def test_feature_checkout_empty_prefix(self):
        self.assertArgparseError('must not by empty',
                                 runGitFlow, 'feature', 'checkout', '')

    @copy_from_fixture('sample_repo')
    def test_feature_checkout_prefix(self):
        runGitFlow('feature', 'checkout', 'rec')

    #--- feature pull ---

    @remote_clone_from_fixture('sample_repo')
    def test_feature_pull(self):
        GitFlow('.').init()
        runGitFlow('feature', 'pull', 'my-remote', 'even')

    @remote_clone_from_fixture('sample_repo')
    def test_feature_pull_empty_remote_raises_error(self):
        GitFlow('.').init()
        self.assertArgparseError('must not by empty',
                                 runGitFlow, 'feature', 'pull', '', 'even')

    @remote_clone_from_fixture('sample_repo')
    def test_feature_pull_nonexisting_remote_raises_error(self):
        GitFlow('.').init()
        self.assertRaises(NoSuchRemoteError,
                          runGitFlow, 'feature', 'pull', 'some-remote', 'even')

    @remote_clone_from_fixture('sample_repo')
    def test_feature_pull_name_is_required(self):
        GitFlow('.').init()
        self.assertRaises(NoSuchBranchError, runGitFlow, 'feature', 'pull', 'my-remote')
        self.assertRaises(NoSuchBranchError, runGitFlow, 'feature', 'pull', 'my-remote', '')


class TestRelease(TestCase):

    @copy_from_fixture('release')
    def test_release_list(self):
        stdout = runGitFlow('release', 'list', capture=1)
        expected = [
          '  1.0'
          ]
        self.assertEqual(stdout.splitlines(), expected)

    @copy_from_fixture('release')
    def test_release_list_verbose(self):
        stdout = runGitFlow('release', 'list', '--verbose', capture=1)
        expected = [
          '  1.0 (based on latest devel)'
          ]
        self.assertEqual(stdout.splitlines(), expected)


    #--- release start ---

    @copy_from_fixture('sample_repo')
    def test_release_start(self):
        runGitFlow('release', 'start', '1.1')
        self.assertIn('rel/1.1', Repo().branches)

    @copy_from_fixture('sample_repo')
    def test_release_start_alt_base(self):
        runGitFlow('release', 'start', '1.1', 'devel')
        self.assertIn('rel/1.1', Repo().branches)

    @copy_from_fixture('sample_repo')
    def test_release_start_wrong_alt_base_raises_error(self):
        self.repo.refs['stable'].checkout()
        fake_commit(self.repo, 'A fake commit on stable')
        self.assertRaises(BaseNotOnBranch,
                          runGitFlow, 'release', 'start', 'wow', 'stable')

    @copy_from_fixture('release')
    def test_release_start_empty_name(self):
        self.assertArgparseError('must not by empty',
                                 runGitFlow, 'release', 'start', '')

    @copy_from_fixture('release')
    def test_release_start_no_name(self):
        self.assertArgparseError('too few arguments',
                                 runGitFlow, 'release', 'start')

    #--- release finish ---

    @copy_from_fixture('release')
    def test_release_finish(self):
        gitflow = GitFlow('.').init()
        gitflow.checkout('release', '1.0')
        runGitFlow('release', 'finish', '1.0')
        self.assertNotIn('rel/1.0', Repo().branches)

    @copy_from_fixture('release')
    def test_release_finish_current(self):
        gitflow = GitFlow('.').init()
        gitflow.checkout('release', '1.0')
        runGitFlow('release', 'finish')
        self.assertNotIn('rel/1.0', Repo().branches)

    @copy_from_fixture('release')
    def test_release_finish_empty_prefix(self):
        gitflow = GitFlow('.').init()
        gitflow.checkout('release', '1.0')
        runGitFlow('release', 'finish', '')
        self.assertNotIn('rel/1.0', Repo().branches)

    @copy_from_fixture('release')
    def test_release_finish_prefix(self):
        gitflow = GitFlow('.').init()
        gitflow.checkout('release', '1.0')
        # release finish requires a name, not a prefix
        self.assertRaises(NoSuchBranchError,
                          runGitFlow, 'release', 'finish', '1')

    #:todo: test-cases for `release finish --rebase`
    #:todo: test-cases for `release finish --fetch`
    #:todo: test-cases for `release finish --keep`
    #:todo: test-cases for `release finish --force-delete`

    #--- release publish ---

    @remote_clone_from_fixture('sample_repo')
    def test_release_publish(self):
        GitFlow('.').init()
        runGitFlow('release', 'start', '1.1')
        runGitFlow('release', 'publish', '1.1')
        self.assertIn('rel/1.1', Repo().remotes['my-remote'].refs)

    @remote_clone_from_fixture('sample_repo')
    def test_release_publish_current(self):
        GitFlow('.').init()
        runGitFlow('release', 'start', '1.1')
        runGitFlow('release', 'publish')
        self.assertIn('rel/1.1', Repo().remotes['my-remote'].refs)

    @remote_clone_from_fixture('sample_repo')
    def test_release_publish_empty_prefix(self):
        GitFlow('.').init()
        runGitFlow('release', 'start', '1.1')
        runGitFlow('release', 'publish', '')
        self.assertIn('rel/1.1', Repo().remotes['my-remote'].refs)

    @remote_clone_from_fixture('sample_repo')
    def test_release_publish_prefix(self):
        GitFlow('.').init()
        runGitFlow('release', 'start', '1.1')
        self.assertRaises(NoSuchBranchError,
                          runGitFlow, 'release', 'publish', '1')

    #--- release track ---

    @remote_clone_from_fixture('release')
    def test_release_track(self):
        GitFlow('.').init()
        runGitFlow('release', 'track', '1.0')

    @remote_clone_from_fixture('sample_repo')
    def test_release_track_name_is_required(self):
        GitFlow('.').init()
        self.assertArgparseError('too few arguments',
                                 runGitFlow, 'release', 'track')
        self.assertArgparseError('must not by empty',
                                 runGitFlow, 'release', 'track', '')

    #--- unsupported `release` subcommands ---

    def test_release_diff_is_no_valid_subcommand(self):
        self.assertArgparseError("invalid choice: 'diff'",
                                 runGitFlow, 'release', 'diff')

    def test_release_rebase_is_no_valid_subcommand(self):
        self.assertArgparseError("invalid choice: 'rebase'",
                                 runGitFlow, 'release', 'rebase')

    def test_release_checkout_is_no_valid_subcommand(self):
        self.assertArgparseError("invalid choice: 'checkout'",
                                 runGitFlow, 'release', 'checkout')

    def test_release_pull_is_no_valid_subcommand(self):
        self.assertArgparseError("invalid choice: 'pull'",
                                 runGitFlow, 'release', 'pull')


class TestHotfix(TestCase):

    @copy_from_fixture('sample_repo')
    def test_hotfix_list(self):
        self.repo.create_head('hf/2.3', 'HEAD')
        stdout = runGitFlow('hotfix', 'list', capture=1)
        expected = [
          '  2.3'
          ]
        self.assertEqual(stdout.splitlines(), expected)

    @copy_from_fixture('sample_repo')
    def test_hotfix_list_verbose(self):
        self.repo.create_head('hf/2.3', 'HEAD')
        stdout = runGitFlow('hotfix', 'list', '--verbose', capture=1)
        expected = [
          '  2.3 (based on latest stable)'
          ]
        self.assertEqual(stdout.splitlines(), expected)

    @copy_from_fixture('release')
    def test_hotfix_list_verbose_tagged(self):
        runGitFlow('release', 'finish', '1.0')
        runGitFlow('hotfix', 'start', '1.0.1')
        fake_commit(self.repo, 'Hotfix commit.')
        stdout = runGitFlow('hotfix', 'list', '--verbose', capture=1)
        expected = [
          '* 1.0.1 (based on v1.0)'
          ]
        self.assertEqual(stdout.splitlines(), expected)

    #--- hotfix start ---

    @copy_from_fixture('sample_repo')
    def test_hotfix_start(self):
        runGitFlow('hotfix', 'start', 'wow-hotfix')
        self.assertIn('hf/wow-hotfix', Repo().branches)

    @copy_from_fixture('sample_repo')
    def test_hotfix_start_alt_base(self):
        runGitFlow('hotfix', 'start', 'wow-hotfix', 'stable')
        self.assertIn('hf/wow-hotfix', Repo().branches)

    @copy_from_fixture('sample_repo')
    def test_hotfix_start_wrong_alt_base_raises_error(self):
        self.assertRaises(BaseNotOnBranch,
                          runGitFlow, 'hotfix', 'start', 'wow-feature', 'devel')

    @copy_from_fixture('sample_repo')
    def test_hotfix_start_empty_name(self):
        self.assertArgparseError('must not by empty',
                                 runGitFlow, 'hotfix', 'start', '')

    @copy_from_fixture('sample_repo')
    def test_hotfix_start_no_name(self):
        self.assertArgparseError('too few arguments',
                                 runGitFlow, 'hotfix', 'start')

    #--- hotfix finish ---

    @remote_clone_from_fixture('sample_repo')
    def test_hotfix_finish(self):
        GitFlow('.').init()
        runGitFlow('hotfix', 'start', 'fast')
        self.assertIn('hf/fast', Repo().branches)
        runGitFlow('hotfix', 'finish', 'fast')
        self.assertNotIn('hf/fast', Repo().branches)

    @copy_from_fixture('sample_repo')
    def test_hotfix_finish_current(self):
        GitFlow('.').init()
        runGitFlow('hotfix', 'start', 'fast')
        runGitFlow('hotfix', 'finish')
        self.assertNotIn('hf/fast', Repo().branches)

    @copy_from_fixture('sample_repo')
    def test_hotfix_finish_empty_prefix(self):
        GitFlow('.').init()
        runGitFlow('hotfix', 'start', 'fast')
        runGitFlow('hotfix', 'finish', '')
        self.assertNotIn('hf/fast', Repo().branches)

    @copy_from_fixture('sample_repo')
    def test_hotfix_finish_prefix(self):
        GitFlow('.').init()
        runGitFlow('hotfix', 'start', 'fast')
        self.assertRaises(NoSuchBranchError,
                          runGitFlow, 'hotfix', 'finish', 'f')

    #:todo: test-cases for `hotfix finish --rebase`
    #:todo: test-cases for `hotfix finish --fetch`
    #:todo: test-cases for `hotfix finish --keep`
    #:todo: test-cases for `hotfix finish --force-delete`

    #--- hotfix publish ---

    @remote_clone_from_fixture('sample_repo')
    def test_hotfix_publish(self):
        GitFlow('.').init()
        runGitFlow('hotfix', 'start', 'wow')
        runGitFlow('hotfix', 'publish', 'wow')
        self.assertIn('hf/wow', Repo().remotes['my-remote'].refs)

    @remote_clone_from_fixture('sample_repo')
    def test_hotfix_publish_current(self):
        GitFlow('.').init()
        runGitFlow('hotfix', 'start', 'wow')
        runGitFlow('hotfix', 'publish')
        self.assertIn('hf/wow', Repo().remotes['my-remote'].refs)

    @remote_clone_from_fixture('sample_repo')
    def test_hotfix_publish_empty_prefix(self):
        GitFlow('.').init()
        runGitFlow('hotfix', 'start', 'wow')
        runGitFlow('hotfix', 'publish', '')
        self.assertIn('hf/wow', Repo().remotes['my-remote'].refs)

    @remote_clone_from_fixture('sample_repo')
    def test_hotfix_publish_prefix(self):
        GitFlow('.').init()
        runGitFlow('hotfix', 'start', 'wow')
        self.assertRaises(NoSuchBranchError,
                          runGitFlow, 'hotfix', 'publish', 'w')

    #--- unsupported `hotfix` subcommands ---

    def test_hotfix_track_is_no_valid_subcommand(self):
        self.assertArgparseError("invalid choice: 'track'",
                                 runGitFlow, 'hotfix', 'track')

    def test_hotfix_diff_is_no_valid_subcommand(self):
        self.assertArgparseError("invalid choice: 'diff'",
                                 runGitFlow, 'hotfix', 'diff')

    def test_hotfix_rebase_is_no_valid_subcommand(self):
        self.assertArgparseError("invalid choice: 'rebase'",
                                 runGitFlow, 'hotfix', 'rebase')

    def test_hotfix_checkout_is_no_valid_subcommand(self):
        self.assertArgparseError("invalid choice: 'checkout'",
                                 runGitFlow, 'hotfix', 'checkout')

    def test_hotfix_pull_is_no_valid_subcommand(self):
        self.assertArgparseError("invalid choice: 'pull'",
                                 runGitFlow, 'hotfix', 'pull')

class TestSupport(TestCase):

    @copy_from_fixture('sample_repo')
    def test_support_list(self):
        self.repo.create_head('supp/1.0-22', 'HEAD')
        stdout = runGitFlow('support', 'list', capture=1)
        expected = [
          '  1.0-22'
          ]
        self.assertEqual(stdout.splitlines(), expected)

    @copy_from_fixture('sample_repo')
    def test_support_list_verbose(self):
        self.repo.create_head('supp/1.0-22', 'HEAD')
        stdout = runGitFlow('support', 'list', '--verbose', capture=1)
        expected = [
          '  1.0-22 (based on latest stable)'
          ]
        self.assertEqual(stdout.splitlines(), expected)

    @copy_from_fixture('release')
    def test_support_list_verbose_tagged(self):
        runGitFlow('release', 'finish', '1.0')
        runGitFlow('support', 'start', '1.0-22')
        fake_commit(self.repo, 'Support commit.')
        stdout = runGitFlow('support', 'list', '--verbose', capture=1)
        expected = [
          '* 1.0-22 (based on v1.0)'
          ]
        self.assertEqual(stdout.splitlines(), expected)

    #--- support start ---

    @copy_from_fixture('sample_repo')
    def test_support_start(self):
        runGitFlow('support', 'start', 'wow-support')
        self.assertIn('supp/wow-support', Repo().branches)

    @copy_from_fixture('sample_repo')
    def test_support_start_alt_base(self):
        runGitFlow('support', 'start', 'wow-support', 'stable')
        self.assertIn('supp/wow-support', Repo().branches)

    @copy_from_fixture('sample_repo')
    def test_support_start_wrong_alt_base_raises_error(self):
        self.assertRaises(BaseNotOnBranch,
                          runGitFlow, 'support', 'start', 'wow-support', 'devel')

    @copy_from_fixture('sample_repo')
    def test_support_start_empty_name(self):
        self.assertArgparseError('must not by empty',
                                 runGitFlow, 'support', 'start', '')

    @copy_from_fixture('sample_repo')
    def test_support_start_no_name(self):
        self.assertArgparseError('too few arguments',
                                 runGitFlow, 'support', 'start')


    #--- unsupported `support` subcommands ---

    def test_support_publish_is_no_valid_subcommand(self):
        self.assertArgparseError("invalid choice: 'finish'",
                                 runGitFlow, 'support', 'finish')

    def test_support_track_is_no_valid_subcommand(self):
        self.assertArgparseError("invalid choice: 'track'",
                                 runGitFlow, 'support', 'track')

    def test_support_diff_is_no_valid_subcommand(self):
        self.assertArgparseError("invalid choice: 'diff'",
                                 runGitFlow, 'support', 'diff')

    def test_support_rebase_is_no_valid_subcommand(self):
        self.assertArgparseError("invalid choice: 'rebase'",
                                 runGitFlow, 'support', 'rebase')

    def test_support_checkout_is_no_valid_subcommand(self):
        self.assertArgparseError("invalid choice: 'checkout'",
                                 runGitFlow, 'support', 'checkout')

    def test_support_pull_is_no_valid_subcommand(self):
        self.assertArgparseError("invalid choice: 'pull'",
                                 runGitFlow, 'support', 'pull')



#
# This file is part of `gitflow`.
# Copyright (c) 2010-2011 Vincent Driessen
# Copyright (c) 2012 Hartmut Goebel
# Distributed under a BSD-like license. For full terms see the file LICENSE.txt
#

from unittest2 import TestCase

from git import GitCommandError

from gitflow.core import GitFlow
from gitflow.exceptions import (
    NoSuchRemoteError, MergeError, BranchExistsError, BaseNotOnBranch,
    BranchTypeExistsError, PrefixNotUniqueError, NoSuchBranchError,
    WorkdirIsDirtyError, TagExistsError)
from gitflow.branches import (
    BranchManager, FeatureBranchManager, ReleaseBranchManager,
    HotfixBranchManager, SupportBranchManager)

from tests.helpers import (copy_from_fixture, remote_clone_from_fixture,
                           fake_commit, all_commits, set_gnupg_home)
from tests.helpers.factory import create_git_repo

__copyright__ = "2010-2011 Vincent Driessen; 2012 Hartmut Goebel"
__license__ = "BSD"

class DummyBranchManager(BranchManager):
    """
    A fictious Dummy branch, used to test Branch functionality, but we need
    a concrete class, since Branch is an "abstract" class.
    """
    identifier = 'abc'
    DEFAULT_PREFIX = 'xyz/'


class TestAbstractBranchManager(TestCase):
    def __prep_explicit_prefix(self):
        repo = create_git_repo(self)
        gitflow = GitFlow(repo)
        fb = DummyBranchManager(gitflow, 'xyz-')
        return repo, fb

    def __prep_explicit_empty_prefix(self):
        repo = create_git_repo(self)
        gitflow = GitFlow(repo)
        fb = DummyBranchManager(gitflow, '')
        return repo, fb

    def test_default_prefix(self):
        repo = create_git_repo(self)
        gitflow = GitFlow(repo)
        fb = DummyBranchManager(gitflow)
        self.assertEquals('xyz/', fb.prefix)

    def test_explicit_prefix(self):
        repo, fb = self.__prep_explicit_prefix()
        self.assertEquals('xyz-', fb.prefix)

    def test_explicit_empty_prefix(self):
        repo, fb = self.__prep_explicit_empty_prefix()
        self.assertEquals('', fb.prefix)

    def test_shorten(self):
        repo = create_git_repo(self)
        gitflow = GitFlow(repo)
        fb = DummyBranchManager(gitflow)
        self.assertEquals('foo', fb.shorten('xyz/foo'))
        self.assertEquals('feature/foo', fb.shorten('feature/foo'))

    def test_full_name(self):
        repo = create_git_repo(self)
        gitflow = GitFlow(repo)
        fb = DummyBranchManager(gitflow)
        self.assertEquals('xyz/foo', fb.full_name('foo'))
        self.assertEquals('xyz/feature/foo', fb.full_name('feature/foo'))

    def test_explicit_prefix_shorten(self):
        repo, fb = self.__prep_explicit_prefix()
        self.assertEquals('foo', fb.shorten('xyz-foo'))
        self.assertEquals('xyz/foo', fb.shorten('xyz/foo'))
        self.assertEquals('feature/foo', fb.shorten('feature/foo'))

    def test_explicit_empty_prefix_shorten(self):
        repo, fb = self.__prep_explicit_empty_prefix()
        self.assertEquals('xyz-foo', fb.shorten('xyz-foo'))
        self.assertEquals('xyz/foo', fb.shorten('xyz/foo'))
        self.assertEquals('feature/foo', fb.shorten('feature/foo'))

    def test_explicit_prefix_full_name(self):
        repo, fb = self.__prep_explicit_prefix()
        self.assertEquals('xyz-foo', fb.full_name('foo'))
        self.assertEquals('xyz-xyz/foo', fb.full_name('xyz/foo'))
        self.assertEquals('xyz-feature/foo', fb.full_name('feature/foo'))

    def test_explicit_empty_prefix_full_name(self):
        repo, fb = self.__prep_explicit_empty_prefix()
        self.assertEquals('foo', fb.full_name('foo'))
        self.assertEquals('xyz-foo', fb.full_name('xyz-foo'))
        self.assertEquals('xyz/foo', fb.full_name('xyz/foo'))
        self.assertEquals('feature/foo', fb.full_name('feature/foo'))


class TestFeatureBranchManager(TestCase):

    def test_defined_members(self):
        # must define at least these members
        members = vars(FeatureBranchManager).keys()
        self.assertIn('DEFAULT_PREFIX', members)
        self.assertIn('identifier', members)

    def test_empty_repo_has_no_features(self):
        repo = create_git_repo(self)
        gitflow = GitFlow(repo)
        mgr = FeatureBranchManager(gitflow)
        self.assertItemsEqual([], mgr.list())

    @copy_from_fixture('sample_repo')
    def test_shorten(self):
        gitflow = GitFlow(self.repo)
        fb = FeatureBranchManager(gitflow)
        self.assertEquals('foo', fb.shorten('feat/foo'))
        self.assertEquals('feature/foo', fb.shorten('feature/foo'))

    @copy_from_fixture('sample_repo')
    def test_full_name(self):
        gitflow = GitFlow(self.repo)
        fb = FeatureBranchManager(gitflow)
        self.assertEquals('feat/foo', fb.full_name('foo'))
        self.assertEquals('feat/feature/foo', fb.full_name('feature/foo'))

    @copy_from_fixture('sample_repo')
    def test_list(self):
        gitflow = GitFlow()
        mgr = FeatureBranchManager(gitflow)
        expected = ['feat/even', 'feat/recursion']
        self.assertItemsEqual(expected, [b.name for b in mgr.list()])

    @copy_from_fixture('sample_repo')
    def test_list_without_matching_prefix(self):
        gitflow = GitFlow()
        mgr = FeatureBranchManager(gitflow, 'fb-')
        expected = []
        self.assertItemsEqual(expected, [b.name for b in mgr.list()])


    @copy_from_fixture('sample_repo')
    def test_by_nameprefix(self):
        gitflow = GitFlow()
        mgr = FeatureBranchManager(gitflow)
        self.assertEquals('feat/even', mgr.by_name_prefix('e').name)
        self.assertEquals('feat/recursion', mgr.by_name_prefix('re').name)

    @copy_from_fixture('sample_repo')
    def test_by_nameprefix_not_unique_enough(self):
        gitflow = GitFlow()
        mgr = FeatureBranchManager(gitflow)
        mgr.create('rescue')
        self.assertRaises(PrefixNotUniqueError, mgr.by_name_prefix, 're')
        self.assertRaises(NoSuchBranchError, mgr.by_name_prefix, 'nonexisting')

    #--- create ---

    def test_create_new_feature_branch(self):
        repo = create_git_repo(self)
        gitflow = GitFlow(repo).init()
        mgr = FeatureBranchManager(gitflow)
        self.assertEqual(0, len(mgr.list()))
        new_branch = mgr.create('foo')
        self.assertEqual(1, len(mgr.list()))
        self.assertEqual('feature/foo', mgr.list()[0].name)
        self.assertEqual(new_branch.commit,
                gitflow.repo.branches['develop'].commit)

    @copy_from_fixture('sample_repo')
    def test_create_new_feature_branch_non_default_prefix(self):
        gitflow = GitFlow(self.repo).init()
        mgr = FeatureBranchManager(gitflow)
        new_branch = mgr.create('foo')
        self.assertEqual(new_branch.name, 'feat/foo')
        self.assertIn('feat/foo', [b.name for b in mgr.list()])
        self.assertEqual(new_branch.commit,
                gitflow.repo.branches['devel'].commit)

    @copy_from_fixture('sample_repo')
    def test_create_new_feature_from_alt_base(self):
        gitflow = GitFlow(self.repo)
        mgr = FeatureBranchManager(gitflow)

        new_branch = mgr.create('foo', 'feat/even')
        self.assertEqual(new_branch.commit,
                gitflow.repo.branches['feat/even'].commit)

    def test_feature_branch_origin(self):
        repo = create_git_repo(self)
        gitflow = GitFlow(repo).init()
        mgr = FeatureBranchManager(gitflow)
        new_branch = mgr.create('foobar')
        self.assertEqual(new_branch.commit,
                gitflow.repo.branches['develop'].commit)

    def test_create_existing_feature_branch_raises_error(self):
        repo = create_git_repo(self)
        gitflow = GitFlow(repo).init()
        mgr = FeatureBranchManager(gitflow)
        mgr.create('foo')
        self.assertRaises(BranchExistsError, mgr.create, 'foo')

    def test_create_feature_branch_fetch_without_remote_raises_error(self):
        repo = create_git_repo(self)
        gitflow = GitFlow(repo).init()
        mgr = FeatureBranchManager(gitflow)
        self.assertRaises(NoSuchRemoteError, mgr.create, 'foo', fetch=True)

    @remote_clone_from_fixture('sample_repo')
    def test_create_feature_from_remote_branch(self):
        remote_branch = self.remote.refs['feat/even']
        rfc0 = remote_branch.commit
        gitflow = GitFlow(self.repo).init()
        mgr = FeatureBranchManager(gitflow)
        mgr.create('even')
        branch = self.repo.active_branch
        self.assertEqual(branch.name, 'feat/even')
        self.assertEqual(branch.commit, rfc0)
        # must be a tracking branch
        self.assertTrue(branch.tracking_branch())
        self.assertEqual(branch.tracking_branch().name, 'my-remote/feat/even')

    @remote_clone_from_fixture('sample_repo')
    def test_create_feature_from_remote_branch_with_develop_behind(self):
        # If BranchManager.create() uses `update`, this test-case has
        # to be adopted, since since `update` change the cloned repo.
        rfc0 = self.remote.refs['feat/even'].commit
        rdc0 = self.remote.refs['devel'].commit
        # add a commit to remote develop branch
        self.remote.refs['devel'].checkout()
        change = fake_commit(self.remote, "Yet another develop commit.")

        gitflow = GitFlow(self.repo).init()
        mgr = FeatureBranchManager(gitflow)
        mgr.create('even')
        # must not advance develop nor feat/even
        self.assertEqual(self.repo.refs['feat/even'].commit, rfc0)
        self.assertEqual(self.repo.refs['devel'].commit, rdc0)
        # change must not be in local repo
        self.assertNotIn(change, all_commits(self.repo))

    @remote_clone_from_fixture('sample_repo')
    def test_create_feature_from_remote_branch_behind(self):
        # If BranchManager.create() uses `update`, this test-case has
        # to be adopted, since since `update` change the cloned repo.
        rfc0 = self.remote.refs['feat/even'].commit
        # add a commit to remote feat/even branch
        self.remote.refs['feat/even'].checkout()
        change = fake_commit(self.remote, "Yet another even commit.")

        gitflow = GitFlow(self.repo).init()
        mgr = FeatureBranchManager(gitflow)
        mgr.create('even')
        # does not advance feat/even, since create() uses `fetch`, not `update`
        self.assertEqual(self.repo.refs['feat/even'].commit, rfc0)
        # change must not be in local repo, since create() uses `fetch`, not `update`
        self.assertNotIn(change, all_commits(self.repo))


    @remote_clone_from_fixture('sample_repo')
    def test_create_feature_fetch_from_remote_branch_behind_really_fetches(self):
        rfc0 = self.remote.refs['feat/even'].commit
        # add a commit to remote feat/even branch
        self.remote.refs['feat/even'].checkout()
        change = fake_commit(self.remote, "Yet another even commit.")

        gitflow = GitFlow(self.repo).init()
        mgr = FeatureBranchManager(gitflow)
        mgr.create('even', fetch=True)
        # must not advance feat/even
        self.assertEqual(self.repo.refs['feat/even'].commit, rfc0)
        # change must nor be in local repo
        self.assertNotIn(change, all_commits(self.repo))


    @copy_from_fixture('sample_repo')
    def test_create_feature_changes_active_branch(self):
        gitflow = GitFlow(self.repo)
        mgr = FeatureBranchManager(gitflow)

        self.assertEquals('feat/recursion', self.repo.active_branch.name)
        mgr.create('foo')
        self.assertEquals('feat/foo', self.repo.active_branch.name)

    @copy_from_fixture('dirty_sample_repo')
    def test_create_feature_raises_error_if_local_changes_would_be_overwritten(self):
        gitflow = GitFlow(self.repo)
        mgr = FeatureBranchManager(gitflow)
        self.assertRaisesRegexp(WorkdirIsDirtyError, 'checked in.*not committed',
                mgr.create, 'foo')

    @copy_from_fixture('dirty_sample_repo')
    def test_create_feature_changes_active_branch_even_if_dirty_but_without_conflicts(self):
        gitflow = GitFlow(self.repo)
        mgr = FeatureBranchManager(gitflow)

        # In this fixture, odd.py contains changes that would be overwritten.
        # Since we don't want to test this here, we revert all local changes in
        # odd.py, but leave the local changes in README.txt.  These changes
        # won't be overwritten by the merge, so git-flow should be able to
        # create a new feature branch if Git can do this
        self.repo.index.reset(index=True, working_tree=True, paths=['odd.py'])
        mgr.create('foo')
        self.assertIn('feature/foo', [b.name for b in mgr.iter()])

    # :todo: test-cases for create with base not on develop
    # :todo: test-cases for create with remote base not on develop

    #---- delete ---

    @copy_from_fixture('sample_repo')
    def test_delete_feature_without_commits(self):
        gitflow = GitFlow(self.repo)
        mgr = FeatureBranchManager(gitflow)

        self.assertEquals(2, len(mgr.list()))
        mgr.create('foo')
        gitflow.develop().checkout()
        self.assertEquals(3, len(mgr.list()))
        mgr.delete('foo')
        self.assertEquals(2, len(mgr.list()))
        self.assertNotIn('feat/foo', [b.name for b in self.repo.branches])


    @copy_from_fixture('sample_repo')
    def test_delete_already_merged_feature(self):
        gitflow = GitFlow(self.repo)
        mgr = FeatureBranchManager(gitflow)

        self.assertEquals(2, len(mgr.list()))
        mgr.create('foo')
        fake_commit(self.repo, 'Dummy commit #1')
        fake_commit(self.repo, 'Dummy commit #2')
        mgr.merge('foo', 'devel')

        self.assertEquals(3, len(mgr.list()))
        mgr.delete('foo')
        self.assertEquals(2, len(mgr.list()))
        self.assertNotIn('feat/foo', [b.name for b in mgr.list()])

    @copy_from_fixture('sample_repo')
    def test_delete_feature_with_commits_raises_error(self):
        gitflow = GitFlow(self.repo)
        mgr = FeatureBranchManager(gitflow)

        self.assertEquals(2, len(mgr.list()))
        mgr.create('foo')
        fake_commit(self.repo, 'A commit on the feature branch.', append=False)
        gitflow.develop().checkout()
        self.assertEquals(3, len(mgr.list()))
        self.assertRaisesRegexp(GitCommandError,
                'The branch .* is not fully merged',
                mgr.delete, 'foo')

    @copy_from_fixture('sample_repo')
    def test_delete_feature_with_commits_forcefully(self):
        gitflow = GitFlow(self.repo)
        mgr = FeatureBranchManager(gitflow)

        self.assertEquals(2, len(mgr.list()))
        mgr.create('foo')
        fake_commit(self.repo, 'A commit on the feature branch.', append=False)
        gitflow.develop().checkout()
        self.assertEquals(3, len(mgr.list()))
        mgr.delete('foo', force=True)
        self.assertEquals(2, len(mgr.list()))
        self.assertNotIn('feat/foo', [b.name for b in self.repo.branches])

    @copy_from_fixture('sample_repo')
    def test_delete_current_feature_raises_error(self):
        gitflow = GitFlow(self.repo)
        mgr = FeatureBranchManager(gitflow)
        mgr.create('foo').checkout()
        self.assertRaisesRegexp(GitCommandError,
                'Cannot delete the branch .* which you are currently on',
                mgr.delete, 'foo')

    @copy_from_fixture('sample_repo')
    def test_delete_non_existing_feature_raises_error(self):
        gitflow = GitFlow(self.repo)
        mgr = FeatureBranchManager(gitflow)
        self.assertRaisesRegexp(GitCommandError, 'branch .* not found',
                mgr.delete, 'nonexisting')

    #--- merge ---

    @copy_from_fixture('sample_repo')
    def test_merge_feature_with_multiple_commits(self):
        gitflow = GitFlow(self.repo)
        mgr = FeatureBranchManager(gitflow)

        dc0 = gitflow.develop().commit
        mgr.merge('even', 'devel')
        dc1 = gitflow.develop().commit

        # Assert merge commit has been made
        self.assertEqual(2, len(dc1.parents))
        self.assertEqual(
                "Merge branch 'feat/even' into devel\n",
                dc1.message)

        # Assert develop branch advanced
        self.assertNotEqual(dc0, dc1)
        # Assert the target-branch is active
        self.assertEqual(gitflow.repo.active_branch.name, 'devel')

    @copy_from_fixture('sample_repo')
    def test_merge_feature_with_single_commit(self):
        gitflow = GitFlow(self.repo).init()
        mgr = FeatureBranchManager(gitflow)

        dc0 = gitflow.develop().commit
        mgr.merge('recursion', 'devel')
        dc1 = gitflow.develop().commit

        # Assert no merge commit has been made
        self.assertEqual(1, len(dc1.parents))
        self.assertEqual('Made the definition of odd recursive.\n',
                dc1.message)

        # Assert develop branch advanced
        self.assertNotEqual(dc0, dc1)
        # Assert the target-branch is active
        self.assertEqual(gitflow.repo.active_branch.name, 'devel')

    def test_merge_feature_without_commits(self):
        repo = create_git_repo(self)
        gitflow = GitFlow(repo).init()
        mgr = FeatureBranchManager(gitflow)

        dc0 = gitflow.develop().commit
        mgr.create('newstuff')
        mgr.merge('newstuff', 'develop')
        dc1 = gitflow.develop().commit

        # Assert the develop tip is unchanged by the merge
        self.assertEqual(dc0, dc1)
        # Assert the target-branch is active
        self.assertEqual(gitflow.repo.active_branch.name, 'develop')

    @copy_from_fixture('sample_repo')
    def test_merge_conflict(self):
        gitflow = GitFlow(self.repo)
        mgr = FeatureBranchManager(gitflow)
        mgr.merge('recursion', 'devel')
        self.assertRaises(MergeError,
                          mgr.merge, 'even', 'devel')
        gitflow.git.rm('odd.py')
        gitflow.git.commit('-F.git/MERGE_MSG')


    #--- finish ---

    @copy_from_fixture('sample_repo')
    def test_finish_feature(self):
        gitflow = GitFlow(self.repo)
        mgr = FeatureBranchManager(gitflow)

        mc0 = gitflow.master().commit
        dc0 = gitflow.develop().commit
        mgr.finish('even')
        mc1 = gitflow.master().commit
        dc1 = gitflow.develop().commit

        # Feature finishes don't advance master, but develop
        self.assertEqual(mc0, mc1)
        self.assertNotEqual(dc0, dc1)

        # Finishing removes the feature branch
        self.assertNotIn('feat/even',
                [b.name for b in self.repo.branches])

        # Merge commit message
        self.assertEquals('Finished feature even.\n', dc1.message)

    @copy_from_fixture('sample_repo')
    def test_finish_feature_keep(self):
        gitflow = GitFlow(self.repo)
        mgr = FeatureBranchManager(gitflow)
        mgr.finish('even', keep=True)
        # feature branch still exists
        self.assertIn('feat/even',
                [b.name for b in self.repo.branches])


    @remote_clone_from_fixture('sample_repo')
    def test_finish_feature_on_unpulled_branch_raises_error(self):
        # branch exists on remote but was not pulled prior to finish
        gitflow = GitFlow(self.repo).init()
        mgr = FeatureBranchManager(gitflow)
        self.assertRaises(NoSuchBranchError, mgr.finish, 'even', push=True)


    @remote_clone_from_fixture('sample_repo')
    def test_finish_feature_push(self):
        remote =  GitFlow(self.remote).init()
        gitflow = GitFlow(self.repo).init()

        rmc0 = remote.master().commit
        rdc0 = remote.develop().commit
        mc0 = gitflow.master().commit
        dc0 = gitflow.develop().commit

        mgr = FeatureBranchManager(gitflow)
        mgr.create('even')
        mgr.finish('even', push=True)

        rmc1 = remote.master().commit
        rdc1 = remote.develop().commit
        mc1 = gitflow.master().commit
        dc1 = gitflow.develop().commit

        # Feature finishes don't advance remote master, but remote develop
        self.assertEqual(rmc0, rmc1)
        self.assertNotEqual(rdc0, rdc1)
        self.assertEqual(mc0, mc1)
        self.assertNotEqual(dc0, dc1)

        # local and remote heads must be the same again
        self.assertEqual(rmc1, mc1)
        self.assertEqual(rdc1, dc1)

        # Finishing removes the local and the remote feature branch
        self.assertNotIn('feat/even',
                [b.name for b in self.repo.branches])
        self.assertNotIn('feat/even',
                [b.name for b in self.remote.branches])

        # Merge commit message
        self.assertEquals('Finished feature even.\n', rdc1.message)


    @remote_clone_from_fixture('sample_repo')
    def test_finish_feature_push_keep(self):
        gitflow = GitFlow(self.repo).init()
        mgr = FeatureBranchManager(gitflow)
        mgr.create('even')
        mgr.finish('even', push=True, keep=True)

        # Finishing removes the local and the remote feature branch
        self.assertIn('feat/even',
                [b.name for b in self.repo.branches])
        self.assertIn('feat/even',
                [b.name for b in self.remote.branches])

    @copy_from_fixture('sample_repo')
    def test_finish_feature_merge_conflict(self):
        gitflow = GitFlow(self.repo).init()
        mgr = FeatureBranchManager(gitflow)
        mgr.finish('recursion')
        self.assertRaises(MergeError,
                          mgr.finish, 'even')
        # resolve the conflict
        gitflow.git.rm('odd.py')
        gitflow.git.commit('-F.git/MERGE_MSG')
        # the feature branch is still here
        self.assertIn('feat/even',
                [b.name for b in self.repo.branches])
        mgr.finish('even')
        # now the feature branch is gone
        self.assertNotIn('feat/even',
                [b.name for b in self.repo.branches])

    @copy_from_fixture('sample_repo')
    def test_finish_feature_unresolved_merge_conflict(self):
        gitflow = GitFlow(self.repo).init()
        mgr = FeatureBranchManager(gitflow)
        mgr.finish('recursion')
        self.assertRaises(MergeError,
                          mgr.finish, 'even')
        # do not resolve, but finish again
        self.assertRaises(GitCommandError,
                          mgr.finish, 'even')

    # :todo: test-cases for finish with rebase
    # :todo: test-cases for finish with rebase-conflicts for both develop

class TestReleaseBranchManager(TestCase):

    def test_defined_members(self):
        # must define at least these members
        members = vars(ReleaseBranchManager).keys()
        self.assertIn('DEFAULT_PREFIX', members)
        self.assertIn('identifier', members)

    @copy_from_fixture('sample_repo')
    def test_shorten(self):
        gitflow = GitFlow(self.repo)
        fb = ReleaseBranchManager(gitflow)
        self.assertEquals('foo', fb.shorten('rel/foo'))
        self.assertEquals('release/foo', fb.shorten('release/foo'))

    @copy_from_fixture('sample_repo')
    def test_full_name(self):
        gitflow = GitFlow(self.repo)
        fb = ReleaseBranchManager(gitflow)
        self.assertEquals('rel/foo', fb.full_name('foo'))
        self.assertEquals('rel/release/foo', fb.full_name('release/foo'))

    def test_empty_repo_has_no_releases(self):
        repo = create_git_repo(self)
        gitflow = GitFlow(repo)
        mgr = ReleaseBranchManager(gitflow)
        self.assertItemsEqual([], mgr.list())

    @copy_from_fixture('release')
    def test_list(self):
        gitflow = GitFlow()
        mgr = ReleaseBranchManager(gitflow)
        expected = ['rel/1.0']
        self.assertItemsEqual(expected, [b.name for b in mgr.list()])

    @copy_from_fixture('release')
    def test_list_without_matching_prefix(self):
        gitflow = GitFlow()
        mgr = ReleaseBranchManager(gitflow, 'rel-')
        expected = []
        self.assertItemsEqual(expected, [b.name for b in mgr.list()])

    @copy_from_fixture('release')
    def test_by_nameprefix(self):
        gitflow = GitFlow()
        mgr = ReleaseBranchManager(gitflow)
        self.assertEquals('rel/1.0', mgr.by_name_prefix('1').name)

    @copy_from_fixture('release')
    def test_by_nameprefix_not_unique_enough(self):
        gitflow = GitFlow()
        mgr = ReleaseBranchManager(gitflow)
        # Create branch without manager since manager enforces there
        # is a single release branch.
        self.repo.create_head('rel/1.1', 'HEAD')
        self.assertRaises(PrefixNotUniqueError, mgr.by_name_prefix, '1.')
        self.assertRaises(NoSuchBranchError, mgr.by_name_prefix, 'nonexisting')

    #--- create ---

    def test_create_new_release_branch(self):
        repo = create_git_repo(self)
        gitflow = GitFlow(repo).init()
        mgr = ReleaseBranchManager(gitflow)
        self.assertEqual(0, len(mgr.list()))
        new_branch = mgr.create('3.14-beta5')
        self.assertEqual(1, len(mgr.list()))
        self.assertEqual('release/3.14-beta5', mgr.list()[0].name)
        self.assertEqual(new_branch.commit,
                gitflow.repo.branches['develop'].commit)

    @copy_from_fixture('sample_repo')
    def test_create_new_release_branch_non_default_prefix(self):
        gitflow = GitFlow(self.repo).init()
        mgr = ReleaseBranchManager(gitflow)
        new_branch = mgr.create('3.14-beta5')
        self.assertEqual(new_branch.name, 'rel/3.14-beta5')
        self.assertIn('rel/3.14-beta5', [b.name for b in mgr.list()])
        self.assertEqual(new_branch.commit,
                gitflow.repo.branches['devel'].commit)

    @copy_from_fixture('sample_repo')
    def test_create_new_release_from_alt_base(self):
        gitflow = GitFlow(self.repo)
        mgr = ReleaseBranchManager(gitflow)

        new_branch = mgr.create('1.0',
                'c8b6deac7ef94f078a426d52c0b1fb3e1221133c')  # devel~1
        self.assertEqual(new_branch.commit.hexsha,
                'c8b6deac7ef94f078a426d52c0b1fb3e1221133c')

    @copy_from_fixture('sample_repo')
    def test_create_new_release_from_wrong_alt_base_raises_error(self):
        gitflow = GitFlow(self.repo)
        mgr = ReleaseBranchManager(gitflow)
        self.assertRaises(BaseNotOnBranch,
                          mgr.create, '1.0', 'feat/even')

    @copy_from_fixture('release')
    def test_create_new_release_for_existing_tag_raises_error(self):
        gitflow = GitFlow(self.repo)
        mgr = ReleaseBranchManager(gitflow)
        mgr.finish('1.0', tagging_info={'message':'Tagging 1.0'})
        self.assertRaises(TagExistsError,
                          mgr.create, '1.0')

    def test_release_branch_origin(self):
        repo = create_git_repo(self)
        gitflow = GitFlow(repo).init()
        mgr = ReleaseBranchManager(gitflow)
        new_branch = mgr.create('1.1')
        self.assertEqual(new_branch.commit,
                gitflow.repo.branches['develop'].commit)

    def test_create_existing_release_branch_raises_error(self):
        repo = create_git_repo(self)
        gitflow = GitFlow(repo).init()
        mgr = ReleaseBranchManager(gitflow)
        mgr.create('1.0')
        self.assertRaises(BranchTypeExistsError, mgr.create, '1.0')

    def test_create_release_branch_fetch_without_remote_raises_error(self):
        repo = create_git_repo(self)
        gitflow = GitFlow(repo).init()
        mgr = ReleaseBranchManager(gitflow)
        self.assertRaises(NoSuchRemoteError, mgr.create, 'foo', fetch=True)

    @remote_clone_from_fixture('release')
    def test_create_release_from_remote_branch(self):
        remote_branch = self.remote.refs['rel/1.0']
        rfc0 = remote_branch.commit
        gitflow = GitFlow(self.repo).init()
        mgr = ReleaseBranchManager(gitflow)
        mgr.create('1.0')
        branch = self.repo.active_branch
        self.assertEqual(branch.name, 'rel/1.0')
        self.assertEqual(branch.commit, rfc0)
        # must be a tracking branch
        self.assertTrue(branch.tracking_branch())
        self.assertEqual(branch.tracking_branch().name, 'my-remote/rel/1.0')

    @remote_clone_from_fixture('release')
    def test_create_release_from_remote_branch_with_develop_behind(self):
        # If BranchManager.create() uses `update`, this test-case has
        # to be adopted, since `update` changes the cloned repo.
        rfc0 = self.remote.refs['rel/1.0'].commit
        rdc0 = self.remote.refs['devel'].commit
        # add a commit to remote develop branch
        self.remote.refs['devel'].checkout()
        change = fake_commit(self.remote, "Yet another develop commit.")

        gitflow = GitFlow(self.repo).init()
        mgr = ReleaseBranchManager(gitflow)
        mgr.create('1.0')
        # must not advance develop nor rel/1.0
        self.assertEqual(self.repo.refs['rel/1.0'].commit, rfc0)
        self.assertEqual(self.repo.refs['devel'].commit, rdc0)
        # change must not be in local repo
        self.assertNotIn(change, all_commits(self.repo))

    @remote_clone_from_fixture('release')
    def test_create_release_from_remote_branch_behind(self):
        # If BranchManager.create() uses `update`, this test-case has
        # to be adopted, since since `update` change the cloned repo.
        rfc0 = self.remote.refs['rel/1.0'].commit
        # add a commit to remote rel/1.0 branch
        self.remote.refs['rel/1.0'].checkout()
        change = fake_commit(self.remote, "Yet another 1.0 commit.")

        gitflow = GitFlow(self.repo).init()
        mgr = ReleaseBranchManager(gitflow)
        mgr.create('1.0')
        # does not advance rel/1.0, since create() uses `fetch`, not `update`
        self.assertEqual(self.repo.refs['rel/1.0'].commit, rfc0)
        # change must not be in local repo, since create() uses `fetch`, not `update`
        self.assertNotIn(change, all_commits(self.repo))


    @remote_clone_from_fixture('release')
    def test_create_release_fetch_from_remote_branch_behind_really_fetches(self):
        rfc0 = self.remote.refs['rel/1.0'].commit
        # add a commit to remote rel/1.0 branch
        self.remote.refs['rel/1.0'].checkout()
        change = fake_commit(self.remote, "Yet another 1.0 commit.")

        gitflow = GitFlow(self.repo).init()
        mgr = ReleaseBranchManager(gitflow)
        mgr.create('1.0', fetch=True)
        # must not advance rel/1.0
        self.assertEqual(self.repo.refs['rel/1.0'].commit, rfc0)
        # change must nor be in local repo
        self.assertNotIn(change, all_commits(self.repo))

    def test_create_release_changes_active_branch(self):
        repo = create_git_repo(self)
        gitflow = GitFlow(repo).init()
        mgr = ReleaseBranchManager(gitflow)

        self.assertEquals('develop', repo.active_branch.name)
        mgr.create('1.0')
        self.assertEquals('release/1.0', repo.active_branch.name)

    @copy_from_fixture('dirty_sample_repo')
    def test_create_release_raises_error_if_local_changes_would_be_overwritten(self):
        gitflow = GitFlow(self.repo)
        mgr = ReleaseBranchManager(gitflow)
        self.assertRaisesRegexp(WorkdirIsDirtyError, 'checked in.*not committed',
                mgr.create, '1.0')

    @copy_from_fixture('dirty_sample_repo')
    def test_create_release_changes_active_branch_even_if_dirty_but_without_conflicts(self):
        gitflow = GitFlow(self.repo)
        mgr = ReleaseBranchManager(gitflow)

        # In this fixture, odd.py contains changes that would be overwritten.
        # Since we don't want to test this here, we revert all local changes in
        # odd.py, but leave the local changes in README.txt.  These changes
        # won't be overwritten by the merge, so git-flow should be able to
        # create a new release branch if Git can do this
        self.repo.index.reset(index=True, working_tree=True, paths=['odd.py'])
        mgr.create('1.0')
        self.assertIn('release/1.0', [b.name for b in mgr.iter()])

    # :todo: test-cases for create with base not on develop
    # :todo: test-cases for create with remote base not on develop

    #---- delete ---

    @copy_from_fixture('sample_repo')
    def test_delete_release_without_commits(self):
        gitflow = GitFlow(self.repo)
        mgr = ReleaseBranchManager(gitflow)

        self.assertEquals(0, len(mgr.list()))
        mgr.create('1.0')
        gitflow.develop().checkout()
        self.assertEquals(1, len(mgr.list()))
        mgr.delete('1.0')
        self.assertEquals(0, len(mgr.list()))
        self.assertNotIn('rel/1.0', [b.name for b in mgr.list()])

    @copy_from_fixture('sample_repo')
    def test_delete_already_merged_release(self):
        gitflow = GitFlow(self.repo)
        mgr = ReleaseBranchManager(gitflow)

        self.assertEquals(0, len(mgr.list()))
        mgr.create('0.7')
        fake_commit(self.repo, 'Dummy commit #1')
        fake_commit(self.repo, 'Dummy commit #2')
        mgr.merge('0.7', 'devel')

        self.assertEquals(1, len(mgr.list()))
        mgr.delete('0.7')
        self.assertEquals(0, len(mgr.list()))
        self.assertNotIn('rel/0.7', [b.name for b in mgr.list()])

    @copy_from_fixture('sample_repo')
    def test_delete_release_with_commits_raises_error(self):
        gitflow = GitFlow(self.repo)
        mgr = ReleaseBranchManager(gitflow)

        self.assertEquals(0, len(mgr.list()))
        mgr.create('0.7')
        fake_commit(self.repo, 'A commit on the release branch.', append=False)
        gitflow.develop().checkout()
        self.assertEquals(1, len(mgr.list()))
        self.assertRaisesRegexp(GitCommandError,
                'The branch .* is not fully merged',
                mgr.delete, '0.7')

    @copy_from_fixture('sample_repo')
    def test_delete_release_with_commits_forcefully(self):
        gitflow = GitFlow(self.repo)
        mgr = ReleaseBranchManager(gitflow)

        self.assertEquals(0, len(mgr.list()))
        mgr.create('0.7')
        fake_commit(self.repo, 'A commit on the release branch.', append=False)
        gitflow.develop().checkout()
        self.assertEquals(1, len(mgr.list()))
        mgr.delete('0.7', force=True)
        self.assertEquals(0, len(mgr.list()))
        self.assertNotIn('rel/0.7', [b.name for b in self.repo.branches])

    @copy_from_fixture('sample_repo')
    def test_delete_current_release_raises_error(self):
        gitflow = GitFlow(self.repo)
        mgr = ReleaseBranchManager(gitflow)
        mgr.create('1.0').checkout()
        self.assertRaisesRegexp(GitCommandError,
                'Cannot delete the branch .* which you are currently on',
                mgr.delete, '1.0')

    @copy_from_fixture('sample_repo')
    def test_delete_non_existing_release_raises_error(self):
        gitflow = GitFlow(self.repo)
        mgr = ReleaseBranchManager(gitflow)
        self.assertRaisesRegexp(GitCommandError, 'branch .* not found',
                mgr.delete, 'nonexisting')

    #--- merge ---
    # No need to test merge for ReleaseBranchManager since it is the
    # same as for FeatureBranch Manager

    #--- finish ---

    @copy_from_fixture('release')
    def test_finish_release(self):
        gitflow = GitFlow(self.repo)
        mgr = ReleaseBranchManager(gitflow)

        mc0 = gitflow.master().commit
        dc0 = gitflow.develop().commit
        mgr.finish('1.0')
        mc1 = gitflow.master().commit
        dc1 = gitflow.develop().commit

        # Release finishes advance both master and develop
        self.assertNotEqual(mc0, mc1)
        self.assertNotEqual(dc0, dc1)
        # master is merged back to develop
        self.assertIn(mc1, dc1.parents)

        # Finishing removes the release branch
        self.assertNotIn('rel/1.0',
                [b.name for b in self.repo.branches])

        # Merge commit message
        self.assertEquals('Finished release 1.0.\n', dc1.message)
        self.assertEquals('Finished release 1.0.\n', mc1.message)


    @copy_from_fixture('release')
    def test_finish_release_keep(self):
        gitflow = GitFlow(self.repo)
        mgr = ReleaseBranchManager(gitflow)
        mgr.finish('1.0', keep=True)
        # release branch still exists
        self.assertIn('rel/1.0',
                [b.name for b in self.repo.branches])

    @copy_from_fixture('release')
    def test_finish_release_tag(self):
        gitflow = GitFlow(self.repo)
        mgr = ReleaseBranchManager(gitflow)
        taginfo = dict(
            message = 'Tagging version 1.0'
            )
        mgr.finish('1.0', tagging_info=taginfo)
        mc1 = gitflow.master().commit
        dc1 = gitflow.develop().commit
        # master is merged back to develop
        self.assertIn(mc1, dc1.parents)
        # tag exists
        self.assertIn('v1.0', self.repo.tags)
        self.assertEqual(self.repo.tags['v1.0'].commit, mc1)
        # tag message
        self.assertEqual(self.repo.tags['v1.0'].tag.message,
                         'Tagging version 1.0')

    @set_gnupg_home
    @copy_from_fixture('release')
    def test_finish_release_tag_sign(self):
        gitflow = GitFlow(self.repo)
        mgr = ReleaseBranchManager(gitflow)
        taginfo = dict(
            message = 'Tagging version 1.0',
            signingkey = 'Dummy Key for Gitflow testing',
            )
        mgr.finish('1.0', tagging_info=taginfo)
        # tag message
        tag = self.repo.tags['v1.0'].tag
        self.assertIn('-----BEGIN PGP SIGNATURE-----', tag.message)


    @remote_clone_from_fixture('release')
    def test_finish_release_on_unpulled_branch_raises_error(self):
        # branch exists on remote but was not pulled prior to finish
        gitflow = GitFlow(self.repo).init()
        mgr = ReleaseBranchManager(gitflow)
        self.assertRaises(NoSuchBranchError, mgr.finish, '1.0', push=True)


    @remote_clone_from_fixture('release')
    def test_finish_release_push(self):
        remote =  GitFlow(self.remote).init()
        # Since remote is no bare repo, checkout some branch untouched
        # by this operation. :fixme: find better solution
        self.remote.heads['feat/even'].checkout()
        gitflow = GitFlow(self.repo).init()

        rmc0 = remote.master().commit
        rdc0 = remote.develop().commit
        mc0 = gitflow.master().commit
        dc0 = gitflow.develop().commit

        mgr = ReleaseBranchManager(gitflow)
        mgr.create('1.0')
        mgr.finish('1.0', push=True)
        rmc1 = remote.master().commit
        rdc1 = remote.develop().commit
        mc1 = gitflow.master().commit
        dc1 = gitflow.develop().commit

        # Release finishes advances master and develop both local and remote
        self.assertNotEqual(rmc0, rmc1)
        self.assertNotEqual(rdc0, rdc1)
        self.assertNotEqual(mc0, mc1)
        self.assertNotEqual(dc0, dc1)

        # local and remote heads must be the same again
        self.assertEqual(rmc1, mc1)
        self.assertEqual(rdc1, dc1)

        # Finishing removes the local and the remote release branch
        self.assertNotIn('rel/1.0',
                [b.name for b in self.repo.branches])
        self.assertNotIn('rel/1.0',
                [b.name for b in self.remote.branches])


        # Merge commit message
        self.assertEquals('Finished release 1.0.\n', rdc1.message)


    @remote_clone_from_fixture('release')
    def test_finish_release_push_keep(self):
        # Since remote is no bare repo, checkout some branch untouched
        # by this operation. :fixme: find better solution
        self.remote.heads['feat/even'].checkout()
        gitflow = GitFlow(self.repo).init()
        mgr = ReleaseBranchManager(gitflow)
        mgr.create('1.0')
        mgr.finish('1.0', push=True, keep=True)

        # release branch still exists local and remote
        self.assertIn('rel/1.0',
                [b.name for b in self.repo.branches])
        self.assertIn('rel/1.0',
                [b.name for b in self.remote.branches])


    @remote_clone_from_fixture('release')
    def test_finish_release_tag_push(self):
        # Since remote is no bare repo, checkout some branch untouched
        # by this operation. :fixme: find better solution
        self.remote.heads['feat/even'].checkout()
        gitflow = GitFlow(self.repo).init()

        mgr = ReleaseBranchManager(gitflow)
        mgr.create('1.0')
        taginfo = dict(
            message = 'Tagging version 1.0'
            )
        mgr.finish('1.0', push=True, tagging_info=taginfo)
        mc1 = gitflow.master().commit
        # remote tag exists
        self.assertIn('v1.0', self.remote.tags)
        self.assertEqual(self.remote.tags['v1.0'].commit, mc1)
        # tag message
        self.assertEqual(self.remote.tags['v1.0'].tag.message,
                         'Tagging version 1.0')


    @set_gnupg_home
    @remote_clone_from_fixture('release')
    def test_finish_release_tag_sign_push(self):
        # Since remote is no bare repo, checkout some branch untouched
        # by this operation. :fixme: find better solution
        self.remote.heads['feat/even'].checkout()
        gitflow = GitFlow(self.repo).init()

        mgr = ReleaseBranchManager(gitflow)
        mgr.create('1.0')
        taginfo = dict(
            message = 'Tagging version 1.0',
            signingkey = 'Dummy Key for Gitflow testing',
            )
        mgr.finish('1.0', push=True, tagging_info=taginfo)
        # tag message
        tag = self.remote.tags['v1.0'].tag
        self.assertIn('-----BEGIN PGP SIGNATURE-----', tag.message)


    def test_finish_release_rebase(self):
        repo = create_git_repo(self)
        gitflow = GitFlow(repo).init()
        mgr = ReleaseBranchManager(gitflow)
        mgr.create('1.0')
        self.assertRaisesRegexp(
            AssertionError, "does not make any sense",
            mgr.finish, '1.0', rebase=True)


    @copy_from_fixture('release')
    def test_finish_release_merge_conflict(self):
        gitflow = GitFlow(self.repo).init()
        fmgr = FeatureBranchManager(gitflow)
        fmgr.finish('even')
        fake_commit(self.repo, 'Overwrite version', filename='VERSION')

        mgr = ReleaseBranchManager(gitflow)
        self.assertRaises(MergeError,
                          mgr.finish, '1.0')
        # resolve the conflict
        gitflow.git.rm('VERSION')
        gitflow.git.commit('-F.git/MERGE_MSG')
        # the release branch is still here
        self.assertIn('rel/1.0',
                [b.name for b in self.repo.branches])
        mgr.finish('1.0')
        # now the release branch is gone
        self.assertNotIn('rel/1.0',
                [b.name for b in self.repo.branches])

    @copy_from_fixture('release')
    def test_finish_release_unresolved_merge_conflict(self):
        gitflow = GitFlow(self.repo).init()
        fmgr = FeatureBranchManager(gitflow)
        fmgr.finish('even')
        fake_commit(self.repo, 'Overwrite version', filename='VERSION')

        mgr = ReleaseBranchManager(gitflow)
        self.assertRaises(MergeError,
                          mgr.finish, '1.0')
        # do not resolve, but finish again
        self.assertRaises(GitCommandError,
                          mgr.finish, '1.0')

    # :todo: test-cases for finish with merge-conflicts on master
    # :todo: test-cases for finish + tag with merge-conflicts on develop
    # :todo: test-cases for finish + tag with merge-conflicts on master

class TestHotfixBranchManager(TestCase):

    def test_defined_members(self):
        members = vars(HotfixBranchManager).keys()
        self.assertEqual(members,
                         ['DEFAULT_PREFIX', '__module__',
                          'identifier', 'default_base', '__doc__'])

    @copy_from_fixture('sample_repo')
    def test_shorten(self):
        gitflow = GitFlow(self.repo)
        fb = HotfixBranchManager(gitflow)
        self.assertEquals('foo', fb.shorten('hf/foo'))
        self.assertEquals('hotfix/foo', fb.shorten('hotfix/foo'))

    @copy_from_fixture('sample_repo')
    def test_full_name(self):
        gitflow = GitFlow(self.repo)
        fb = HotfixBranchManager(gitflow)
        self.assertEquals('hf/foo', fb.full_name('foo'))
        self.assertEquals('hf/hotfix/foo', fb.full_name('hotfix/foo'))

    def test_empty_repo_has_no_hotfixes(self):
        repo = create_git_repo(self)
        gitflow = GitFlow(repo)
        mgr = HotfixBranchManager(gitflow)
        self.assertItemsEqual([], mgr.list())

    def test_create_new_hotfix_branch(self):
        repo = create_git_repo(self)
        gitflow = GitFlow(repo).init()
        mgr = HotfixBranchManager(gitflow)
        self.assertEqual(0, len(mgr.list()))
        new_branch = mgr.create('1.2.3')
        self.assertEqual(1, len(mgr.list()))
        self.assertEqual('hotfix/1.2.3', mgr.list()[0].name)
        self.assertEqual(new_branch.commit,
                gitflow.repo.branches['master'].commit)

    @copy_from_fixture('sample_repo')
    def test_create_new_hotfix_branch_non_default_prefix(self):
        gitflow = GitFlow(self.repo).init()
        mgr = HotfixBranchManager(gitflow)
        new_branch = mgr.create('1.2.3')
        self.assertEqual('hf/1.2.3', mgr.list()[0].name)
        self.assertEqual(new_branch.commit,
                gitflow.repo.branches['stable'].commit)

    # :todo: test-cases for create with base not on master
    # :todo: test-cases for create with remote base not on master

    @copy_from_fixture('sample_repo')
    def test_hotfix_branch_origin(self):
        gitflow = GitFlow()
        mgr = HotfixBranchManager(gitflow)
        new_branch = mgr.create('3.14-beta5')
        self.assertEqual(new_branch.commit,
                gitflow.repo.branches['stable'].commit)

    @copy_from_fixture('sample_repo')
    def test_finish_hotfix(self):
        gitflow = GitFlow(self.repo)
        mgr = HotfixBranchManager(gitflow)
        mgr.create('1.2.3')
        fake_commit(self.repo, 'Bogus commit')
        fake_commit(self.repo, 'Foo commit')
        fake_commit(self.repo, 'Fake commit')
        fake_commit(self.repo, 'Dummy commit')

        mc0 = gitflow.master().commit
        dc0 = gitflow.develop().commit
        mgr.finish('1.2.3')
        mc1 = gitflow.master().commit
        dc1 = gitflow.develop().commit

        # Hotfix finishes advance both master and develop
        self.assertNotEqual(mc0, mc1)
        self.assertNotEqual(dc0, dc1)

        # Finishing removes the hotfix branch
        self.assertNotIn('hf/1.2.3',
                [b.name for b in self.repo.branches])

        # Merge commit message
        self.assertEquals('Finished hotfix 1.2.3.\n', dc1.message)


class TestSupportBranchManager(TestCase):

    def test_defined_members(self):
        # must define at least these members
        members = vars(SupportBranchManager).keys()
        self.assertIn('DEFAULT_PREFIX', members)
        self.assertIn('identifier', members)
        self.assertIn('default_base',  members)

    @copy_from_fixture('sample_repo')
    def test_shorten(self):
        gitflow = GitFlow(self.repo)
        fb = SupportBranchManager(gitflow)
        self.assertEquals('foo', fb.shorten('supp/foo'))
        self.assertEquals('support/foo', fb.shorten('support/foo'))

    @copy_from_fixture('sample_repo')
    def test_full_name(self):
        gitflow = GitFlow(self.repo)
        fb = SupportBranchManager(gitflow)
        self.assertEquals('supp/foo', fb.full_name('foo'))
        self.assertEquals('supp/support/foo', fb.full_name('support/foo'))

    def test_empty_repo_has_no_support(self):
        repo = create_git_repo(self)
        gitflow = GitFlow(repo)
        mgr = SupportBranchManager(gitflow)
        self.assertItemsEqual([], mgr.list())

    def test_create_new_support_branch(self):
        repo = create_git_repo(self)
        gitflow = GitFlow(repo).init()
        mgr = SupportBranchManager(gitflow)
        self.assertEqual(0, len(mgr.list()))
        new_branch = mgr.create('1.x')
        self.assertEqual(1, len(mgr.list()))
        self.assertEqual('support/1.x', mgr.list()[0].name)
        self.assertEqual(new_branch.commit,
                gitflow.repo.branches['master'].commit)

    @copy_from_fixture('sample_repo')
    def test_create_new_support_branch_non_default_prefix(self):
        gitflow = GitFlow(self.repo).init()
        mgr = SupportBranchManager(gitflow)
        new_branch = mgr.create('1.x')
        self.assertEqual('supp/1.x', mgr.list()[0].name)
        self.assertEqual(new_branch.commit,
                gitflow.repo.branches['stable'].commit)

    @copy_from_fixture('sample_repo')
    def test_support_branch_origin(self):
        gitflow = GitFlow()
        mgr = SupportBranchManager(gitflow)
        new_branch = mgr.create('legacy')
        self.assertEqual(new_branch.commit,
                gitflow.repo.branches['stable'].commit)

    # :todo: test-cases for create with base not on master
    # :todo: test-cases for create with remote base not on master

    def test_support_branches_cannot_be_finished(self):
        repo = create_git_repo(self)
        gitflow = GitFlow(repo).init()
        mgr = SupportBranchManager(gitflow)
        mgr.create('1.x')
        self.assertRaises(NotImplementedError, mgr.finish, '1.x')




#
# This file is part of `gitflow`.
# Copyright (c) 2010-2011 Vincent Driessen
# Copyright (c) 2012 Hartmut Goebel
# Distributed under a BSD-like license. For full terms see the file LICENSE.txt
#

from unittest2 import TestCase, skip

from gitflow.core import GitFlow
from tests.helpers.factory import create_git_repo

__copyright__ = "2010-2011 Vincent Driessen; 2012 Hartmut Goebel"
__license__ = "BSD"


class TestGitFlow(TestCase):

    @skip
    def test_create(self):
        create_git_repo(self)
        gitflow = GitFlow()

        self.assertEquals([], self.repo.branches)

        gitflow.init()

        self.assertEquals(['develop', 'master'],
                [b.name for b in self.repo.branches])

        fb = gitflow.create('feature', 'foo')





# -*- encoding: utf-8 -*-
import unittest2
import datetime
from textwrap import dedent

from busyflow.pivotal import PivotalClient


class TestPivotalClient(unittest2.TestCase):

    def test_make_story_xml(self):
        client = PivotalClient('')
        result = client.stories.make_story_xml('Some story', 'Do some stuff', 'feature',
                                               requested_by=u'Ignas Mikalajūnas')
        expected = """
        <story>
            <name>Some story</name>
            <description>Do some stuff</description>
            <requested_by>Ignas Mikalajūnas</requested_by>
            <story_type>feature</story_type>
        </story>"""
        self.assertEqual(result, dedent(expected).strip())

        result = client.stories.make_story_xml('Some story', 'Do some stuff', 'feature',
                                               requested_by=u'Ignas Mikalajūnas',
                                               estimate=3,
                                               current_state='started',
                                               labels=['l1', 'l2'])
        expected = """
        <story>
            <name>Some story</name>
            <description>Do some stuff</description>
            <requested_by>Ignas Mikalajūnas</requested_by>
            <story_type>feature</story_type>
            <estimate type="integer">3</estimate>
            <current_state>started</current_state>
            <labels>l1,l2</labels>
        </story>"""
        self.assertEqual(result, dedent(expected).strip())


        result = client.stories.make_story_xml('Some story', 'Do some stuff', 'feature',
                                               requested_by=u'Ignas Mikalajūnas',
                                               estimate=3,
                                               current_state='started',
                                               labels=[])
        expected = """
        <story>
            <name>Some story</name>
            <description>Do some stuff</description>
            <requested_by>Ignas Mikalajūnas</requested_by>
            <story_type>feature</story_type>
            <estimate type="integer">3</estimate>
            <current_state>started</current_state>
            <labels>,</labels>
        </story>"""
        self.assertEqual(result, dedent(expected).strip())


    def test_parse_story_xml(self):
        client = PivotalClient('')
        story_xml = """
          <story>
            <id type="integer">123</id>
            <project_id type="integer">456</project_id>
            <story_type>feature</story_type>
            <url>http://www.pivotaltracker.com/story/show/123</url>
            <estimate type="integer">1</estimate>
            <current_state>accepted</current_state>
            <description></description>
            <name>Purchase Order</name>
            <requested_by>Anon</requested_by>
            <owned_by>Anon</owned_by>
            <created_at type="datetime">2012/01/01 11:00:00 UTC</created_at>
            <updated_at type="datetime">2012/01/01 11:00:00 EST</updated_at>
            <accepted_at type="datetime">2012/01/01 11:00:00 EEST</accepted_at>
          </story>
        """
        story = client.parseContent(story_xml)
        self.assertEqual(
            story,
            {u'story': {u'current_state': u'accepted',
                        u'description': '',
                        u'estimate': 1,
                        u'id': 123,
                        u'name': u'Purchase Order',
                        u'owned_by': u'Anon',
                        u'project_id': 456,
                        u'requested_by': u'Anon',
                        u'story_type': u'feature',
                        u'created_at': datetime.datetime(2012, 1, 1, 11, 0),
                        u'updated_at': datetime.datetime(2012, 1, 1, 6, 0),
                        u'accepted_at': datetime.datetime(2012, 1, 1, 14, 0),
                        u'url': u'http://www.pivotaltracker.com/story/show/123'}})




""" Extensions and utility functions for Python Review Board API.
"""

import urllib
from .rb import Api20Client, make_rbclient
from operator import itemgetter


def is_story_approved(rb_server_url, story_id, auth=None):
    if not auth:
        auth = {'username': '', 'password': ''}
    rb_api = make_rbclient(rb_server_url, auth['username'], auth['password'])

    reviews_for_branch = rb_api.get_review_requests(branch=story_id)

    def is_shipited(review_request):
        # Get all the reviews for review requestr with id @review_request.
        reviews = rb_api.get_reviews_for_review_request(
            review_request['id'])
        # Return True if the review has been approved.
        return any(r['ship_it'] for r in reviews)

    return (len(reviews_for_branch) > 0 and
        all(is_shipited(r) for r in reviews_for_branch))

def get_latest_review_request_for_branch(rb_server_url, branch):
    """ Returns the most recently created review request in Reviewboard
        for a given branch (i.e., the request with the greatest id).
    """
    # Let the username and password be read from the cookie file.
    rb_api = make_rbclient(rb_server_url, '', '')
    reviews_for_branch = rb_api.get_review_requests(options=None, branch=branch)
    if reviews_for_branch:
        return max(reviews_for_branch, key=itemgetter('id'))
    else:
        return None


def get_reviews_for_review_request(self, rev_req_id):
    rsp = self._api_request(
        'GET', '/api/review-requests/%s/reviews' % rev_req_id)
    return rsp['reviews']

Api20Client.get_reviews_for_review_request = get_reviews_for_review_request


def get_review_requests(self, options=None, branch=None):
    options = options or {}
    rsp = self._api_request(
        'GET', '/api/review-requests?%s' % urllib.urlencode(options))
    requests = rsp['review_requests']
    if branch:
        return [r for r in requests if r['branch'] == branch]
    else:
        return requests

Api20Client.get_review_requests = get_review_requests



#
# This file is part of `gitflow`.
# Copyright (c) 2010-2011 Vincent Driessen
# Copyright (c) 2012 Hartmut Goebel
# Distributed under a BSD-like license. For full terms see the file LICENSE.txt
#

import os
import shutil
import tempfile
from functools import wraps

from unittest2 import TestCase
from git import Repo

__copyright__ = "2010-2011 Vincent Driessen; 2012 Hartmut Goebel"
__license__ = "BSD"

def sandboxed(f):
    """
    This decorator sets up a temporary, self-destructing empty directory and
    switches the current directory to it.  The name of the directory is stored
    in self.sandbox attribute.  Files created/modified outside of the sandbox
    aren't cleaned up by this method.
    """
    @wraps(f)
    def _inner(self, *args, **kwargs):
        assert isinstance(self, TestCase)
        ram_disk = '/Volumes/RAM_Disk'
        dir = None
        if os.path.exists(ram_disk):
            dir = ram_disk
        self.sandbox = tempfile.mkdtemp(dir=dir)
        self.addCleanup(shutil.rmtree, self.sandbox)

        os.chdir(self.sandbox)

        # Call the function
        f(self, *args, **kwargs)
    return _inner


def set_gnupg_home(func):
    """
    Decorator which changes the current working dir to the one of the
    git repository in order to assure relative paths are handled
    correctly.
    """
    @wraps(func)
    def _inner(*args, **kwargs):
        root = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
        gpghome = os.path.join(root, 'fixtures', 'gnupg')
        oldval = os.environ.get('GNUPGHOME')
        os.environ['GNUPGHOME'] = gpghome
        func(*args, **kwargs)
        if oldval is not None:
            os.environ['GNUPGHOME'] = oldval
        else:
            del os.environ['GNUPGHOME']
    return _inner


def copy_from_fixture(fixture_name):
    """
    This decorator sets up a temporary, self-destructing sandbox and copies
    a given fixture into it.  The repo is accessible inside the function via the
    self.repo attribute.  This is useful for fixtures that represent changes in
    the configuration or dirty working directories.
    """
    def _outer(f):
        @wraps(f)
        @sandboxed
        def _inner(self, *args, **kwargs):
            root = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
            src = os.path.join(root, 'fixtures', fixture_name)
            dest = os.path.join(self.sandbox, fixture_name)
            shutil.copytree(src, dest)
            os.chdir(dest)
            shutil.move('dot_git', '.git')
            self.repo = Repo(dest)
            f(self, *args, **kwargs)
        return _inner
    return _outer

def clone_from_fixture(fixture_name):
    """
    This decorator sets up a temporary, self-destructing sandbox, cloned from
    a given fixture.  In contrast to a filesystem copy, a clone always has fresh
    configuration and a clean working directory.

    The repo is accesible via the self.repo attribute inside the tests.
    """
    def _outer(f):
        @wraps(f)
        @sandboxed
        def _inner(self, *args, **kwargs):
            root = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
            git_dir = os.path.join(root, 'fixtures', fixture_name, 'dot_git')
            self.repo = Repo(git_dir).clone(self.sandbox)
            f(self, *args, **kwargs)
        return _inner
    return _outer


def copy_gitflow_config(src, dest):
    reader = src.config_reader(config_level='repository')
    writer = dest.config_writer(config_level='repository')
    for section in reader.sections():
        if section.startswith('gitflow '):
            for item, value in reader.items(section):
                writer.set_value(section, item, value)
    del writer

def remote_clone_from_fixture(fixture_name, copy_config=True):
    """
    This decorator sets up a temporary, self-destructing sandbox,
    cloned from a sandboxed copy of the fixture.  In contrast to a
    filesystem copy, the clone always has a clean working directory.

    :param copy_config:
         Copy the `gitflow` parts of the original repo into the clone.

    The repo is accesible via the self.repo attribute inside the
    tests, the remote (clond) repo via `self.remote`.
    """
    def _outer(f):
        @wraps(f)
        @sandboxed
        def _inner(self, *args, **kwargs):
            root = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
            src = os.path.join(root, 'fixtures', fixture_name)
            dest = os.path.join(self.sandbox, 'remote')
            shutil.copytree(src, dest)
            os.chdir(dest)
            shutil.move('dot_git', '.git')
            self.remote = Repo(dest)
            clone = os.path.join(self.sandbox, 'clone')
            self.repo = self.remote.clone(clone, origin='my-remote')
            if copy_config:
                copy_gitflow_config(self.remote, self.repo)
            self.repo.config_writer(config_level='repository').set_value(
                'gitflow', 'origin', 'my-remote')
            os.chdir(clone)
            f(self, *args, **kwargs)
        return _inner
    return _outer


def git_working_dir(func):
    """
    Decorator which changes the current working dir to the one of the
    git repository in order to assure relative paths are handled
    correctly.
    """
    # Adopted from GitPython's git.index.util.git_working_dir
    # Copyright (C) 2008, 2009 Michael Trier (mtrier@gmail.com) and
    #    contributors
    # Released under the BSD License:
    # http://www.opensource.org/licenses/bsd-license.php
    @wraps(func)
    def set_git_working_dir(repo, *args, **kwargs):
        cur_wd = os.getcwd()
        os.chdir(repo.working_tree_dir)
        try:
            return func(repo, *args, **kwargs)
        finally:
            os.chdir(cur_wd)

    return set_git_working_dir


@git_working_dir
def fake_commit(repo, message, append=True, filename='newfile.py'):
    if append:
        f = open(filename, 'a')
    else:
        f = open(filename, 'w')
    try:
        f.write('This is a dummy change.\n')
    finally:
        f.close()
    repo.index.add([filename])
    return repo.index.commit(message)


def all_commits(repo, heads=None):
    s = set([])
    if heads:
        heads = [h for h in repo.heads if h.name in heads]
    else:
        heads = repo.heads
    for h in heads:
        s |= set(repo.iter_commits(h))
    return s



#
# Shamelessly ripped from
# http://code.activestate.com/recipes/576949-find-all-subclasses-of-a-given-class/
#
def itersubclasses(cls, _seen=None):
    """
    itersubclasses(cls)

    Generator over all subclasses of a given class, in depth first order.

    >>> list(itersubclasses(int)) == [bool]
    True
    >>> class A(object): pass
    >>> class B(A): pass
    >>> class C(A): pass
    >>> class D(B,C): pass
    >>> class E(D): pass
    >>> 
    >>> for cls in itersubclasses(A):
    ...     print(cls.__name__)
    B
    D
    E
    C
    >>> # get ALL (new-style) classes currently defined
    >>> [cls.__name__ for cls in itersubclasses(object)] #doctest: +ELLIPSIS
    ['type', ...'tuple', ...]
    """

    if not isinstance(cls, type):
        raise TypeError('itersubclasses must be called with '
                        'new-style classes, not %.100r' % cls)
    if _seen is None: _seen = set()
    try:
        subs = cls.__subclasses__()
    except TypeError: # fails only when cls is type
        subs = cls.__subclasses__(cls)
    for sub in subs:
        if sub not in _seen:
            _seen.add(sub)
            yield sub
            for sub in itersubclasses(sub, _seen):
                yield sub





# -*- coding: utf-8 ; ispell-local-dictionary: "american" -*-
"""
git-flow init
"""
#
# This file is part of `gitflow`.
# Copyright (c) 2010-2011 Vincent Driessen
# Copyright (c) 2012 Hartmut Goebel
# Distributed under a BSD-like license. For full terms see the file LICENSE.txt
#

import sys
import re

try:
    # this will trigger readline functionality for raw_input
    import readline
except:
    # readline is optional and may not be available on all installations
    pass

from gitflow.core import GitFlow as CoreGitFlow, warn, info

from gitflow.exceptions import (AlreadyInitialized, NotInitialized,
                                NoSuchLocalBranchError, NoSuchBranchError,
                                IllegalCommunicationProtocol, RemoteNotFound)

import gitflow.review as rb
import gitflow.pivotal as pt

__copyright__ = "2010-2011 Vincent Driessen; 2012 Hartmut Goebel"
__license__ = "BSD"


def _ensure_SSH():
    gitflow_origin = gitflow.get('gitflow.origin')
    try:
        origin_url = gitflow.get('remote.' + gitflow_origin + '.url')
    except Exception:
        raise RemoteNotFound(
                "Please configure git remote '{0}' before proceesing further." \
                        .format(gitflow_origin))
    if re.match('(https?://|git://)', origin_url):
        raise IllegalCommunicationProtocol("Only SSH remotes are supported.")

class GitFlow(CoreGitFlow):

    @staticmethod
    def _has_configured(branch_func):
        try:
            branch_func()
        except (NotInitialized, IndexError):
            return False
        return True

    def has_master_configured(self):
        return self._has_configured(self.master)

    def has_develop_configured(self):
        return self._has_configured(self.develop)

    def get_default(self, setting):
        return self.get(setting, self.defaults[setting])
        

def _ask_branch(args, name, desc1, desc2, suggestions, filter=[]):
    # Two cases are distinguished:
    # 1. A fresh git repo (without any branches)
    #    We will create a new master/develop branch for the user
    # 2. Some branches do already exist
    #    We will disallow creation of new master/develop branches and
    #    rather allow to use existing branches for git-flow.
    name = 'gitflow.branch.' + name
    default_name = gitflow.get_default(name)
    local_branches = [b
                      for b in gitflow.branch_names()
                      if b not in filter]
    if not local_branches:
        if not filter:
            print "No branches exist yet. Base branches must be created now."
        should_check_existence = False
        default_suggestion = default_name
    else:
        should_check_existence = True
        print
        print "Which branch should be used for %s?" % desc1
        for b in local_branches:
            print '  -', b
        for default_suggestion in [default_name] + suggestions:
            if default_suggestion in local_branches:
                break
        else:
            default_suggestion = ''

    if args.use_defaults and default_suggestion:
        print "Branch name for %s:" % desc2, default_suggestion
        branch_name = default_suggestion
    else:
        answer = raw_input("Branch name for %s: [%s] "
                           % (desc2, default_suggestion))
        branch_name = answer.strip() or default_suggestion
    if not branch_name:
        raise SystemExit('You need to give a branch name.')
    # check existence in case of an already existing repo
    if branch_name in filter:
        raise SystemExit("Production and integration branches should differ.")
    if should_check_existence:
        # if no local branch exists and a remote branch of the same
        # name exists, checkout that branch and use it for the local branch
        if not branch_name in local_branches:
            remote_name = gitflow.origin_name(branch_name)
            if remote_name in gitflow.branch_names(remote=True):
                branch = gitflow.repo.create_head(branch_name, remote_name)
                info("Created local branch %s based on %s."
                     % (branch_name, remote_name))
            else:
                raise NoSuchLocalBranchError(branch_name)

    # store the name of the develop branch
    gitflow.set(name, branch_name)
    return branch_name


def _ask_config(args, name, question):
    default_suggestion = gitflow.get_default(name)
    if args.use_defaults:
        print question +':', default_suggestion
        answer = default_suggestion
    else:
        answer = raw_input(question + '? [' + default_suggestion + '] ')
        answer = answer.strip() or default_suggestion
        if answer == '-':
            answer = ''
    gitflow.set(name, answer)

def _ask_prefix(args, name, question):
    name = 'gitflow.prefix.' + name
    if not gitflow.get(name, None) or args.force:
        _ask_config(args, name, question)

def _ask_name(args, name, question):
    name = 'gitflow.' + name
    if not gitflow.get(name, None) or args.force:
        _ask_config(args, name, question)

def _pick(title, source):
    print
    print 'Please choose one of the following %s:' % title
    msg = '    Loading...'
    sys.stdout.write(msg)
    sys.stdout.flush()
    suggestions = source()
    sys.stdout.write('\r' * len(msg))
    while True:
        i = 0
        for sid, sname in suggestions:
            i += 1
            print '    [%d] %s' % (i, sname)
        a = raw_input("Insert the sequence number (or 'q' to quit): ")
        if a == 'q':
            raise SystemExit('Operation canceled.')
        try:
            answer = int(a)
        except ValueError:
            print 'Please specify a number betweet 1 and %i:' % i
            continue
        if answer >= 1 and answer <= i:
            return suggestions[answer-1]

def _ask_pt_projid():
    gitflow.set('gitflow.pt.projectid',
                _pick('Pivotal Tracker projects', pt.list_projects)[0])

def _ask_rb_repoid():
    gitflow.set('gitflow.rb.repoid',
                _pick('Review Board repositories', rb.list_repos)[0])

def run_default(args):
    global gitflow
    gitflow = GitFlow()
    gitflow._enforce_git_repo()
    gitflow._enforce_services()

    if gitflow.is_initialized():
        if not args.force:
            raise AlreadyInitialized()

    if args.use_defaults:
        warn("Using default branch names.")

    _ask_name(args, "origin", "Remote name to use as origin in git flow")
 
    # Make sure that origin uses SSH protocol for communication,
    # otherwise Review Board is going to fail.
    _ensure_SSH()

    #-- add a master branch if no such branch exists yet
    if gitflow.has_master_configured() and not args.force:
        master_branch = gitflow.master_name()
    else:
        master_branch = _ask_branch(args,
            'master',
            'bringing forth production releases',
            'production releases',
            ['production', 'main', 'master'])

    #-- add a develop branch if no such branch exists yet
    if gitflow.has_develop_configured() and not args.force:
        develop_branch = gitflow.develop_name()
    else:
        develop_branch = _ask_branch(args,
            'develop',
            'integration of the "next release"',
            '"next release" development',
            ['develop', 'int', 'integration', 'master'],
            filter=[master_branch])

    if not gitflow.is_initialized() or args.force:
        print
        print "How to name your supporting branch prefixes?"

    _ask_prefix(args, "feature", "Feature branches")
    _ask_prefix(args, "release", "Release branches")
    _ask_prefix(args, "hotfix", "Hotfix branches")
    _ask_prefix(args, "support", "Support branches")
    _ask_prefix(args, "versiontag", "Version tag prefix")

    _ask_name(args, 'release.versionmatcher',
            'Regular expression for matching release numbers')

    _ask_pt_projid()
    _ask_rb_repoid()

    # assert the gitflow repo has been correctly initialized
    assert gitflow.is_initialized()

    gitflow.init(master_branch, develop_branch)



# api code for the reviewboard extension, inspired/copied from reviewboard
# post-review code.

import cookielib
import getpass
import mimetools
import os
import urllib2
import simplejson
#import mercurial.ui
from urlparse import urljoin, urlparse

class APIError(Exception):
    pass

class ReviewBoardError(Exception):
    def __init__(self, json=None):
        self.msg = None
        self.code = None
        self.tags = {}

        if isinstance(json, str) or isinstance(json, unicode):
            try:
                json = simplejson.loads(json)
            except:
                self.msg = "Unknown error - non-JSON response"
                return

        if json:
            if json.has_key('err'):
                self.msg = json['err']['msg']
                self.code = json['err']['code']
            for key, value in json.items():
                if isinstance(value,unicode) or isinstance(value,str) or \
                    key == 'fields':
                    self.tags[key] = value

    def __str__(self):
        if self.msg:
            return ("%s (%s)" % (self.msg, self.code)) + \
                ''.join([("\n%s: %s" % (k, v)) for k,v in self.tags.items()])
        else:
            return Exception.__str__(self)

class Repository:
    """
    Represents a ReviewBoard repository
    """
    def __init__(self, id, name, tool, path):
        self.id = id
        self.name = name
        self.tool = tool
        self.path = path

class ReviewBoardHTTPPasswordMgr(urllib2.HTTPPasswordMgr):
    """
    Adds HTTP authentication support for URLs.

    Python 2.4's password manager has a bug in http authentication when the
    target server uses a non-standard port.  This works around that bug on
    Python 2.4 installs. This also allows post-review to prompt for passwords
    in a consistent way.

    See: http://bugs.python.org/issue974757
    """
    def __init__(self, reviewboard_url):
        self.passwd  = {}
        self.rb_url  = reviewboard_url
        self.rb_user = None
        self.rb_pass = None

    def set_credentials(self, username, password):
        self.rb_user = username
        self.rb_pass = password

    def find_user_password(self, realm, uri):
        if uri.startswith(self.rb_url):
            if self.rb_user is None or self.rb_pass is None:
                print "==> HTTP Authentication Required"
                print 'Enter username and password for "%s" at %s' % \
                    (realm, urlparse(uri)[1])
                print "Username:"
                self.rb_user = raw_input()
                # Changed by TOM
                #self.rb_user = mercurial.ui.ui().prompt('Username: ')
                self.rb_pass = getpass.getpass('Password: ')

            return self.rb_user, self.rb_pass
        else:
            # If this is an auth request for some other domain (since HTTP
            # handlers are global), fall back to standard password management.
            return urllib2.HTTPPasswordMgr.find_user_password(self, realm, uri)

class ApiRequest(urllib2.Request):
    """
    Allows HTTP methods other than GET and POST to be used
    """
    def __init__(self, method, *args, **kwargs):
        self._method = method
        urllib2.Request.__init__(self, *args, **kwargs)

    def get_method(self):
        return self._method

class HttpErrorHandler(urllib2.HTTPDefaultErrorHandler):
    """
    Error handler that doesn't throw an exception for any code below 400.
    This is necessary because RB returns 2xx codes other than 200 to indicate
    success.
    """
    def http_error_default(self, req, fp, code, msg, hdrs):
        if code >= 400:
            return urllib2.HTTPDefaultErrorHandler.http_error_default(self,
                req, fp, code, msg, hdrs)
        else:
            result = urllib2.HTTPError( req.get_full_url(), code, msg, hdrs, fp)
            result.status = code
            return result

class HttpClient:
    def __init__(self, url, proxy=None):
        if not url.endswith('/'):
            url = url + '/'
        self.url       = url
        if 'APPDATA' in os.environ:
            homepath = os.environ["APPDATA"]
        elif 'USERPROFILE' in os.environ:
            homepath = os.path.join(os.environ["USERPROFILE"], "Local Settings",
                                    "Application Data")
        elif 'HOME' in os.environ:
            homepath = os.environ["HOME"]
        else:
            homepath = ''
        self.cookie_file = os.path.join(homepath, ".post-review-cookies.txt")
        self._cj = cookielib.MozillaCookieJar(self.cookie_file)
        self._password_mgr = ReviewBoardHTTPPasswordMgr(self.url)
        self._opener = opener = urllib2.build_opener(
                        urllib2.ProxyHandler(proxy),
                        urllib2.UnknownHandler(),
                        urllib2.HTTPHandler(),
                        HttpErrorHandler(),
                        urllib2.HTTPErrorProcessor(),
                        urllib2.HTTPCookieProcessor(self._cj),
                        urllib2.HTTPBasicAuthHandler(self._password_mgr),
                        urllib2.HTTPDigestAuthHandler(self._password_mgr)
                        )
        urllib2.install_opener(self._opener)

    def set_credentials(self, username, password):
        self._password_mgr.set_credentials(username, password)

    def api_request(self, method, url, fields=None, files=None):
        """
        Performs an API call using an HTTP request at the specified path.
        """
        try:
            rsp = self._http_request(method, url, fields, files)
            if rsp:
                return self._process_json(rsp)
            else:
                return None
        except APIError, e:
            rsp, = e.args
            raise ReviewBoardError(rsp)

    def has_valid_cookie(self):
        """
        Load the user's cookie file and see if they have a valid
        'rbsessionid' cookie for the current Review Board server.  Returns
        true if so and false otherwise.
        """
        try:
            parsed_url = urlparse(self.url)
            host = parsed_url[1]
            path = parsed_url[2] or '/'

            # Cookie files don't store port numbers, unfortunately, so
            # get rid of the port number if it's present.
            host = host.split(":")[0]

            #print("Looking for '%s %s' cookie in %s" % \
            #      (host, path, self.cookie_file))
            self._cj.load(self.cookie_file, ignore_expires=True)

            try:
                cookie = self._cj._cookies[host][path]['rbsessionid']

                if not cookie.is_expired():
            #        print("Loaded valid cookie -- no login required")
                    return True

            #    print("Cookie file loaded, but cookie has expired")
            except KeyError:
                print("Cookie file loaded, but no cookie for this server")
        except IOError, error:
            print("Couldn't load cookie file: %s" % error)

        return False

    def _http_request(self, method, path, fields, files):
        """
        Performs an HTTP request on the specified path.
        """
        if path.startswith('/'):
            path = path[1:]
        url = urljoin(self.url, path)
        body = None
        headers = {}
        if fields or files:
            content_type, body = self._encode_multipart_formdata(fields, files)
            headers = {
                'Content-Type': content_type,
                'Content-Length': str(len(body))
                }

        try:
            if type(url) == unicode:
                url = url.encode('utf8')
            r = ApiRequest(method, url, body, headers)
            data = urllib2.urlopen(r).read()
            self._cj.save(self.cookie_file)
            return data
        except urllib2.URLError, e:
            if not hasattr(e, 'code'):
                raise
            if e.code >= 400:
                raise
            else:
                return ""
        except urllib2.HTTPError, e:
            raise ReviewBoardError(e.read())

    def _process_json(self, data):
        """
        Loads in a JSON file and returns the data if successful. On failure,
        APIError is raised.
        """
        rsp = simplejson.loads(data)

        if rsp['stat'] == 'fail':
            raise APIError, rsp

        return rsp

    def _encode_multipart_formdata(self, fields, files):
        """
        Encodes data for use in an HTTP POST.
        """
        BOUNDARY = mimetools.choose_boundary()
        content = ""

        fields = fields or {}
        files = files or {}

        for key in fields:
            content += "--" + BOUNDARY + "\r\n"
            content += "Content-Disposition: form-data; name=\"%s\"\r\n" % key
            content += "\r\n"
            content += str(fields[key]) + "\r\n"

        for key in files:
            filename = files[key]['filename']
            value = files[key]['content']
            content += "--" + BOUNDARY + "\r\n"
            content += "Content-Disposition: form-data; name=\"%s\"; " % key
            content += "filename=\"%s\"\r\n" % filename
            content += "\r\n"
            content += value + "\r\n"

        content += "--" + BOUNDARY + "--\r\n"
        content += "\r\n"

        content_type = "multipart/form-data; boundary=%s" % BOUNDARY

        return content_type, content

class ApiClient:
    def __init__(self, httpclient):
        self._httpclient = httpclient

    def _api_request(self, method, url, fields=None, files=None):
        return self._httpclient.api_request(method, url, fields, files)

class Api20Client(ApiClient):
    """
    Implements the 2.0 version of the API
    """

    def __init__(self, httpclient):
        ApiClient.__init__(self, httpclient)
        self._repositories = None
        self._requestcache = {}

    def login(self, username=None, password=None):
        self._httpclient.set_credentials(username, password)
        return

    def repositories(self):
        if not self._repositories:
            rsp = self._api_request('GET', '/api/repositories/?max-results=500')
            self._repositories = [Repository(r['id'], r['name'], r['tool'],
                                             r['path'])
                                  for r in rsp['repositories']]
        return self._repositories

    def new_request(self, repo_id, fields={}, diff='', parentdiff=''):
        req = self._create_request(repo_id)
        self._set_request_details(req, fields, diff, parentdiff)
        self._requestcache[req['id']] = req
        return req['id']

    def update_request(self, id, fields={}, diff='', parentdiff='', publish=True):
        req = self._get_request(id)
        self._set_request_details(req, fields, diff, parentdiff)
        if publish:
            self.publish(id)

    def publish(self, id):
        req = self._get_request(id)
        drafturl = req['links']['draft']['href']
        self._api_request('PUT', drafturl, {'public':'1'})

    def _create_request(self, repo_id):
        data = { 'repository': repo_id }
        result = self._api_request('POST', '/api/review-requests/', data)
        return result['review_request']

    def _get_request(self, id):
        if self._requestcache.has_key(id):
            return self._requestcache[id]
        else:
            result = self._api_request('GET', '/api/review-requests/%s/' % id)
            self._requestcache[id] = result['review_request']
            return result['review_request']

    def _set_request_details(self, req, fields, diff, parentdiff):
        drafturl = req['links']['self']['href']
        if fields:
            if fields.get("status"):
                key = 'self'
            else:
                key = 'draft'
            drafturl = req['links'][key]['href']
            self._api_request('PUT', drafturl, fields)
        if diff:
            diffurl = req['links']['diffs']['href']
            data = {'path': {'filename': 'diff', 'content': diff}}
            if parentdiff:
                data['parent_diff_path'] = \
                    {'filename': 'parent_diff', 'content': parentdiff}
            self._api_request('POST', diffurl, {}, data)

class Api10Client(ApiClient):
    """
    Implements the 1.0 version of the API
    """

    def __init__(self, httpclient):
        ApiClient.__init__(self, httpclient)
        self._repositories = None
        self._requests = None

    def _api_post(self, url, fields=None, files=None):
        return self._api_request('POST', url, fields, files)

    def login(self, username=None, password=None):
        if not username and not password and self._httpclient.has_valid_cookie():
            return

        if not username:
            print "Username:"
            username = raw_input()
            # Changed by TOM
            #username = mercurial.ui.ui().prompt('Username: ')
        if not password:
            password = getpass.getpass('Password: ')

        self._api_post('/api/json/accounts/login/', {
            'username': username,
            'password': password,
        })

    def repositories(self):
        if not self._repositories:
            rsp = self._api_post('/api/json/repositories/')
            self._repositories = [Repository(r['id'], r['name'], r['tool'],
                                             r['path'])
                                  for r in rsp['repositories']]
        return self._repositories

    def requests(self):
        if not self._requests:
            rsp = self._api_post('/api/json/reviewrequests/all/')
            self._requests = rsp['review_requests']
        return self._requests

    def new_request(self, repo_id, fields={}, diff='', parentdiff=''):
        repository_path = None
        for r in self.repositories():
            if r.id == int(repo_id):
                repository_path = r.path
                break
        if not repository_path:
            raise ReviewBoardError, ("can't find repository with id: %s" % \
                                        repo_id)

        id = self._create_request(repository_path)

        self._set_request_details(id, fields, diff, parentdiff)

        return id

    def update_request(self, id, fields={}, diff='', parentdiff='', publish=True):
        request_id = None
        for r in self.requests():
            if r['id'] == int(id):
                request_id = int(id)
                break
        if not request_id:
            raise ReviewBoardError, ("can't find request with id: %s" % id)

        self._set_request_details(request_id, fields, diff, parentdiff)

        return request_id

    def publish(self, id):
        self._api_post('api/json/reviewrequests/%s/publish/' % id)

    def _create_request(self, repository_path):
        data = { 'repository_path': repository_path }
        rsp = self._api_post('/api/json/reviewrequests/new/', data)

        return rsp['review_request']['id']

    def _set_request_field(self, id, field, value):
        self._api_post('/api/json/reviewrequests/%s/draft/set/' %
                                id, { field: value })

    def _upload_diff(self, id, diff, parentdiff=""):
        data = {'path': {'filename': 'diff', 'content': diff}}
        if parentdiff:
            data['parent_diff_path'] = \
                {'filename': 'parent_diff', 'content': parentdiff}
        rsp = self._api_post('/api/json/reviewrequests/%s/diff/new/' % \
                                id, {}, data)

    def _set_fields(self, id, fields={}):
        for field in fields:
            self._set_request_field(id, field, fields[field])

    def _set_request_details(self, id, fields, diff, parentdiff):
        self._set_fields(id, fields)
        if diff:
            self._upload_diff(id, diff, parentdiff)
        if fields or diff:
            self._save_draft(id)

    def _save_draft(self, id):
        self._api_post("/api/json/reviewrequests/%s/draft/save/" % id )

def make_rbclient(url, username, password, proxy=None, apiver=''):
    httpclient = HttpClient(url, proxy)

    if not httpclient.has_valid_cookie():
        if not username:
            print "Username:"
            username = raw_input()
            # Changed by TOM
            #username = mercurial.ui.ui().prompt('Username: ')
        if not password:
            password = getpass.getpass('Password: ')
        httpclient.set_credentials(username, password)

    if not apiver:
        # Figure out whether the server supports API version 2.0
        try:
            httpclient.api_request('GET', '/api/')
            apiver = '2.0'
        except:
            apiver = '1.0'

    if apiver == '2.0':
        return Api20Client(httpclient)
    elif apiver == '1.0':
        cli = Api10Client(httpclient)
        cli.login(username, password)
        return cli
    else:
        raise Exception("Unknown API version: %s" % apiver)




#
# This file is part of `gitflow`.
# Copyright (c) 2010-2011 Vincent Driessen
# Copyright (c) 2012 Hartmut Goebel
# Distributed under a BSD-like license. For full terms see the file LICENSE.txt
#

import os
import shutil
import tempfile

from unittest2 import TestCase
from git import Repo

__copyright__ = "2010-2011 Vincent Driessen; 2012 Hartmut Goebel"
__license__ = "BSD"

def create_sandbox(testcase):
    """
    Sets up a temporary, self-destructing directory and chdir to it.
    """
    assert isinstance(testcase, TestCase)

    dir = None
    ram_disk = '/Volumes/RAM_Disk'
    if os.path.exists(ram_disk):
        dir = ram_disk
    sandbox = tempfile.mkdtemp(dir=dir)
    testcase.addCleanup(shutil.rmtree, sandbox)

    os.chdir(sandbox)

    return sandbox


def create_git_repo(testcase):
    """
    Sets up a temporary, self-destructing empty Git repository using "git init".
    There hasn't been any git flow initialization yet.
    """
    create_sandbox(testcase)
    repo = Repo.init()
    return repo




import time
import pkg_resources
import datetime
from poster.encode import multipart_encode
from poster.encode import MultipartParam
from dateutil import parser
import httplib2
import urllib
import logging

from xmlbuilder import XMLBuilder
from xml.dom import minidom
from xml.parsers.expat import ExpatError

log = logging.getLogger(__name__)


def error_from_response(http_response, error_class,
                        response_body=None, parsed_body=None):
    error = error_class('%s: %i, %s' % (http_response.reason, http_response.status,
                                        parsed_body if parsed_body else response_body))
    error.status = http_response.status
    error.reason = http_response.reason
    error.body = response_body
    error.parsed_body = parsed_body
    error.headers = http_response.items()
    return error


class Error(Exception):
    pass


class RequestError(Error):
    """Base http exception"""
    status = None
    reason = None
    body = None
    headers = None
    parsed_body = None


class UnauthorizedError(RequestError):
    """Unauthorized exception"""


def parse(node):
    # determine type
    if node.attributes:
        obj_type = node.attributes["type"].value
    elif node.nodeName in ["stories", "notes"]:
        obj_type = "array"
    elif node.nodeName in ["labels"]:
        obj_type = "csv"
    elif len(node.childNodes) > 1:
        obj_type = "dictionary"
    else:
        obj_type = "string"
    return NODE_PARSERS.get(obj_type, parse_dict)(node)


def parse_csv(node):
    if len(node.childNodes) == 0:
        return []
    value = node.childNodes[0].wholeText.strip()
    return value.split(",")


_tzmap = None
def get_tzmap():
    global _tzmap
    if _tzmap is None:
        _tzmap = {}
        tzmap_file = pkg_resources.resource_string('gitflow.busyflow.pivotal', 'tzmap.txt')
        for line in tzmap_file.splitlines():
            [short_name, long_name, offset] = line.split('\t')
            _tzmap[short_name] = offset
    return _tzmap


def parse_string_to_dt(value):
    try:
        parts = value.split(' ')
        dt = ' '.join(parts[:-1])
        tz = parts[-1]
        offset = get_tzmap().get(tz, '')

        # Parser returns different times for UTC+3000 and +3000 suffix
        # The former will add 3 hours to the given time, while the latter
        # will subtract it
        offset = offset.replace('UTC', '')

        time_tuple = parser.parse(dt + ' ' + offset).utctimetuple()
    except ValueError:
        time_tuple = time.strptime(dt, '%Y/%m/%d %H:%M:%S')
    return datetime.datetime(
        time_tuple.tm_year,
        time_tuple.tm_mon,
        time_tuple.tm_mday,
        time_tuple.tm_hour,
        time_tuple.tm_min,
        time_tuple.tm_sec)


def parse_datetime(node):
    if len(node.childNodes) == 0:
        return None
    value = node.childNodes[0].wholeText.strip()
    return parse_string_to_dt(value)


def parse_list(parent_node):
    new_list = []
    for child_node in parent_node.childNodes:
        if child_node.nodeName != "#text":
            value = parse(child_node)
            new_list.append(value)
    return new_list


def parse_dict(parent_node):
    new_dict = {}
    for child_node in parent_node.childNodes:
        if child_node.nodeName != "#text":
            value = parse(child_node)
            new_dict[child_node.nodeName] = value
    return new_dict


def parse_string(node):
    if not node.childNodes:
        return ""
    return node.childNodes[0].wholeText.strip()


NODE_PARSERS = {'array': parse_list,
                'string': parse_string,
                'dictionary': parse_dict,
                'integer': lambda node: int(node.childNodes[0].wholeText.strip()),
                'csv': parse_csv,
                'datetime': parse_datetime}


class Endpoint(object):
    def __init__(self, pivotal):
        self.pivotal = pivotal

    def _get(self, endpoint, **params):
        return self.pivotal._apicall(endpoint, 'GET', **params)

    def _post(self, endpoint, **params):
        return self.pivotal._apicall(endpoint, 'POST', **params)

    def _put(self, endpoint, **params):
        return self.pivotal._apicall(endpoint, 'PUT', **params)

    def _delete(self, endpoint, **params):
        return self.pivotal._apicall(endpoint, 'DELETE', **params)


class ProjectEndpoint(Endpoint):

    def get(self, project_id):
        return self._get("projects/%s" % project_id)

    def all(self):
        return self._get("projects")

    def activities(self, project_id, limit=None, occurred_since_date=None, newer_than_version=None):
        return self._get("projects/%s/activities" % project_id, limit=limit,
                         occurred_since_date=occurred_since_date,
                         newer_than_version=newer_than_version)

    def make_post_xml(self, name, iteration_length, point_scale):
        project = XMLBuilder(format=True)
        with project.project:
            if name is not None:
                project << ('name', name)
            if iteration_length is not None:
                project << ('iteration_length', iteration_length)
            if point_scale is not None:
                project << ('point_scale', point_scale)
        return str(project)

    def post(self, name, iteration_length, point_scale):
        # XXX actually test it, add missing attributes
        body = self.make_project_xml(name, iteration_length, point_scale)
        return self._post("projects", body=body)


class IterationEndpoint(Endpoint):

    def all(self, project_id, limit=None, offset=None):
        return self._get("projects/%s/iterations" % project_id, limit=limit, offset=offset)

    def done(self, project_id, limit=None, offset=None):
        return self._get("projects/%s/iterations/done" % project_id, limit=limit, offset=offset)

    def current(self, project_id, limit=None, offset=None):
        return self._get("projects/%s/iterations/current" % project_id, limit=limit, offset=offset)

    def backlog(self, project_id, limit=None, offset=None):
        return self._get("projects/%s/iterations/backlog" % project_id, limit=limit, offset=offset)

    def current_backlog(self, project_id, limit=None, offset=None):
        return self._get("projects/%s/iterations/current_backlog" % project_id, limit=limit, offset=offset)


class ActivityEndpoint(Endpoint):

    def all(self, limit=None, occurred_since_date=None, newer_than_version=None):
        return self._get("activities", limit=limit,
                         occurred_since_date=occurred_since_date,
                         newer_than_version=newer_than_version)


class TokenEndpoint(Endpoint):

    def active(self, username, password):
        return self._post('tokens/active',
                          username=username,
                          password=password)


class StoryEndpoint(Endpoint):

    def make_story_xml(self, name=None, description=None, story_type=None,
                       owned_by=None, requested_by=None, estimate=None, current_state=None, labels=None):
        story = XMLBuilder('story')
        if name is not None:
            story.name(name)
        if description is not None:
            story.description(description)
        if requested_by is not None:
            story.requested_by(requested_by)
        if owned_by is not None:
            story.owned_by(owned_by)
        if story_type is not None:
            story.story_type(story_type)
        if estimate is not None:
            story.estimate(str(estimate), type='integer')
        if current_state is not None:
            story.current_state(current_state)
        if labels is not None:
            label_string = ','
            if labels:
                label_string = ','.join(labels)
            story.labels(label_string)

        return str(story)

    def make_comment_xml(self, text, author):
        note = XMLBuilder('note')
        note.text(text)
        if author is not None:
            note.author(author)
        return str(note)

    def all(self, project_id, query=None, limit=None, offset=None):
        return self._get("projects/%s/stories" % project_id, query=query, limit=limit, offset=offset)

    def get(self, project_id, story_id):
        return self._get("projects/%s/stories/%s" % (project_id, story_id))

    def post(self, project_id, name, description, story_type,
             owned_by=None, requested_by=None, estimate=None,
             current_state=None, labels=None):
        body = self.make_story_xml(name, description, story_type,
                                   owned_by=owned_by,
                                   requested_by=requested_by,
                                   estimate=estimate,
                                   current_state=current_state,
                                   labels=labels)
        return self._post("projects/%s/stories" % project_id, body=body)

    def update(self, project_id, story_id,
               name=None, description=None, owned_by=None, requested_by=None,
               story_type=None, estimate=None, current_state=None, labels=None):
        body = self.make_story_xml(name, description, story_type,
                                   owned_by=owned_by,
                                   requested_by=requested_by,
                                   estimate=estimate,
                                   current_state=current_state,
                                   labels=labels)
        return self._put("projects/%s/stories/%s" % (project_id, story_id), body=body)

    def deliver_all_finished_stories(self, project_id):
        return self._put("projects/%s/stories/deliver_all_finished" % project_id)

    def delete(self, project_id, story_id):
        return self._delete("projects/%s/stories/%s" % (project_id, story_id))

    def move(self, project_id, story_id, target_id, move='after'):
        params = {}
        params['move[move]'] = move
        params['move[target]'] = target_id
        return self._post("projects/%s/stories/%s/moves" % (project_id, story_id),
                          **params)

    def add_comment(self, project_id, story_id, text, author=None):
        body = self.make_comment_xml(text, author)
        return self._post("projects/%s/stories/%s/notes" % (project_id, story_id), body=body)

    def add_attachment(self, project_id, story_id,
                       filename,
                       file_obj,
                       filetype,
                       filesize=None):
        if isinstance(file_obj, basestring):
            file_obj = open(file_obj, 'rb')
        file_data = MultipartParam(name='Filedata',
                                   filename=filename,
                                   filetype=filetype,
                                   fileobj=file_obj,
                                   filesize=filesize)

        params = {'Filedata': file_data}
        data, mp_headers = multipart_encode(params)

        if 'Content-Length' in mp_headers:
            mp_headers['Content-Length'] = str(mp_headers['Content-Length'])

        return self._post("projects/%s/stories/%s/attachments" % (project_id, story_id),
                          body="".join(list(data)),
                          headers=mp_headers)


class PivotalClient(object):

    def __init__(self, token,
                 base_url="https://www.pivotaltracker.com/services/v3/",
                 parse_xml=True, cache=None, timeout=None, proxy_info=None):
        self.token = token
        self.base_url = base_url
        self.parse_xml = parse_xml
        self.client = httplib2.Http(cache=cache, timeout=timeout, proxy_info=proxy_info)

        # connect endpoints
        self.projects = ProjectEndpoint(self)
        self.stories = StoryEndpoint(self)
        self.activities = ActivityEndpoint(self)
        self.iterations = IterationEndpoint(self)
        self.tokens = TokenEndpoint(self)

    def _apicall(self, endpoint, method, **params):
        url = '%s%s' % (self.base_url, endpoint)
        body = params.pop('body', '')
        _headers = params.pop('headers', {})
        cleaned_params = dict([(k, v) for k, v in params.iteritems() if v])

        headers = {'X-TrackerToken': self.token}
        if method in ['POST', 'PUT'] and body:
            headers['Content-type'] = 'application/xml'

        headers.update(_headers)

        if cleaned_params:
            assert not body # can't have body and parameters at the same time
            body = urllib.urlencode(cleaned_params)
            if method == 'GET':
                url = '%s?%s' % (url, body)
                body = ''

        resp, content = self.client.request(url, method=method, body=body, headers=headers)

        parsed_content = None
        try:
            parsed_content = self.parseContent(content)
        except (ValueError, ExpatError):
            log.error(resp)
            log.error(content)

        error_cls = RequestError
        if resp.status == 401:
            error_cls = UnauthorizedError

        if resp.status != 200:
            raise error_from_response(resp, error_cls,
                                      content, parsed_content)

        if parsed_content is None:
            # generate the error once more
            self.parseContent(content)

        return parsed_content

    def parseContent(self, content):
        dom = minidom.parseString(content)

        if self.parse_xml:
            return parse_dict(dom)
        else:
            return dom




import re
import subprocess as sub
import difflib as diff
import gitflow.pivotal as pivotal
import reviewboard.extensions as rb_ext
import gitflow.core as core
import sys

from gitflow.core import GitFlow
from gitflow.exceptions import (GitflowError, MultipleReviewRequestsForBranch,
                                NoSuchBranchError)


class ReviewNotAcceptedYet(GitflowError): pass

_gitflow = GitFlow()

def _get_repo_id():
    return _gitflow._safe_get('gitflow.rb.repoid')

def _get_server():
    return _gitflow._safe_get('reviewboard.server')

def _get_url():
    return _gitflow._safe_get('reviewboard.url')

def _get_client():
    return rb_ext.make_rbclient(_get_server(), '', '')

def _get_develop_name():
    return _gitflow.develop_name()

def _get_branch(identifier, name):
    prefix = _gitflow.get_prefix(identifier)
    name = _gitflow.nameprefix_or_current(identifier, name)
    return prefix + name


def list_repos():
    repos = _get_client().repositories()
    return [(r.id, r.name) for r in repos]


class BranchReview(object):
    def __init__(self, branch, rev_range=None):
        assert rev_range is None or len(rev_range) == 2
        self._branch = branch
        self._client = _get_client()
        if rev_range:
            self._rev_range = rev_range

    def __getattr__(self, name):
        if name == '_rid':
            self._rid = self._branch_to_rid(self._branch)
            return self._rid
        raise AttributeError

    def get_id(self):
        return self._rid

    def get_url(self):
        if self._url:
            return self._url
        elif self.get_id():
            return '{0}r/{1}/'.format(_get_url(), self.get_id())
        else:
            raise AttributeError('Neither review id nor review url is defined.')

    def post(self, summary=None, desc=None):
        assert self._rev_range
        cmd = ['rbt', 'post',
               '--branch', self._branch,
               '--revision-range={0[0]}:{0[1]}'.format(self._rev_range)]
        if summary is None or desc is None:
            cmd.append('--guess-fields')
        else:
            cmd.append('--summary=' + str(summary))
            cmd.append('--description=' + str(desc))

        if self._rid:
            sys.stdout.write('updating %s ... ' % str(self._rid))
            cmd.append('--review-request-id')
            cmd.append(str(self._rid))
        else:
            sys.stdout.write('new review ... ')
        output = sub.check_output(cmd)
        print
        print
        print '>>> rbt output'
        print output
        print '<<< rbt output'

        # Use list comprehension to get rid of emply trailing strings.
        self._url = [line for line in output.split('\n') if line != ''][-1]
        self._rid = [f for f in self._url.split('/') if f != ''][-1]

    def submit(self):
        assert self._rid
        if not self.is_accepted():
            raise ReviewNotAcceptedYet('review %s not accepted yet' \
                                       % str(self._rid))
        self._update(status='submitted')

    def is_accepted(self):
        assert self._rid
        reviews = self._client.get_reviews_for_review_request(self._rid)
        return any(r['ship_it'] for r in reviews)

    def _update(self, **kwargs):
        self._client.update_request(self.get_id(), fields=kwargs, publish=True)

    def _branch_to_rid(self, branch):
        options = dict(repository=_get_repo_id())
        reviews = self._client.get_review_requests(options=options,
                                                   branch=self._branch)
        if len(reviews) > 1:
            raise MultipleReviewRequestsForBranch(self._branch)
        elif len(reviews) == 1:
            return reviews[0]['id']

    @classmethod
    def from_prefix(cls, prefix):
        client = _get_client()
        options = dict(repository=_get_repo_id(), status='all')
        reviews = [r for r in client.get_review_requests(options=options)
                     if r['branch'].startswith(prefix)]
        if len(reviews) == 0:
            raise NoSuchBranchError(
                    'No review request found for branch prefixed with ' + prefix)
        elif len(reviews) == 1:
            r = reviews[0]
            t = type('BranchReview', (cls,), dict(_rid=r['id']))
            return t(r['branch'])
        else:
            raise MultipleReviewRequestsForBranch(r['branch'])

    @classmethod
    def from_identifier(cls, identifier, name, rev_range=None):
        prefix = _gitflow.get_prefix(identifier)
        name = _gitflow.nameprefix_or_current(identifier, name)
        return cls(prefix + name, rev_range)


def post_review(self, identifier, name, post_new):
    mgr = self.managers[identifier]
    branch = mgr.by_name_prefix(name)

    sys.stdout.write("Walking the Git reflogs to find review request parent (might "
        "take a couple seconds)...\n")
    sys.stdout.flush()

    if not post_new:
        parent = find_last_patch_parent(self.develop_name(), branch.name)
        if not parent:
            print ("Could not find any merges into %s, using full patch." %
                self.develop_name())
    if not parent:
        parent = get_branch_parent(branch.name)

    if not parent:
        raise GitflowError("Could not find parent for branch '%s'!" %
            branch.name)

    story_id = pivotal.get_story_id_from_branch_name(branch.name)
    story = pivotal.get_story(story_id)

    cmd = ['post-review', '--branch', branch.name,
        '--guess-description',
        '--parent', self.develop_name(),
        '--revision-range', '%s:%s' % (parent, branch.name)]

    if post_new:
        # Create a new request.
        cmd += ['--summary', "'%s'" % story['story']['name']]
    else:
        req = rb_ext.get_latest_review_request_for_branch(
            _gitflow.get('reviewboard.server'), branch.name)
        if req:
            # Update an existing request.
            cmd += ['-r', str(req['id'])]
        else:
            # Create a new request.
            cmd += ['--summary', "'%s'" % story['story']['name']]

    print "Posting a review using command: %s" % ' '.join(cmd)
    proc = sub.Popen(cmd, stdout=sub.PIPE)
    (out, err) = proc.communicate()
    # Post a comment to the relevant Pivotal Tracker story (to make it easier to
    # track review requests).
    review_url = out.strip().split('\n')[-1]
    if not review_url.startswith('http'):
        print ("Could not determine review URL (probably an error when "
            "posting the review")
        return
    if '-r' in cmd:
        comment = "Review request %s updated."
    else:
        comment = "Review request posted: %s"
    pivotal.add_comment_to_story(story_id, comment % review_url)

core.GitFlow.post_review = post_review


def get_branch_parent(branch_name):
    proc = sub.Popen(
        ['git', 'reflog', 'show', branch_name],
        env={'GIT_PAGER': 'cat'}, stdout=sub.PIPE)
    (out, err) = proc.communicate()
    lines = out.strip().split('\n')
    parent = None
    for line in lines:
        parts = line.split(' ')
        if len(parts) >= 3 and (parts[2]).startswith('branch'):
            parent = parts[0]
    if not parent:
        parent = lines[-1].split(' ')[0]
    return parent


def find_last_patch_parent(develop_name, branch_name):
    proc = sub.Popen(
        ['git', 'reflog', 'show', develop_name],
        env={'GIT_PAGER': 'cat'}, stdout=sub.PIPE)
    (out, err) = proc.communicate()
    lines = out.strip().split('\n')
    lines = [l for l in lines if ("merge %s" % branch_name) in l]
    if len(lines) < 2:
        return None
    else:
        # Get the hash of the second most recent merge.
        return lines[1].split(' ')[0]

def get_feature_ancestor(feature):
    develop = _gitflow.develop_name()

    # Get all commits being a part of the respective branches.
    develop_ancestors = sub.check_output(
            ['git', 'rev-list', '--first-parent', develop])
    feature_ancestors = sub.check_output(
            ['git', 'rev-list', '--first-parent', feature])

    # Compute the diff between the lists.
    ancestor_diff = diff.unified_diff(
            develop_ancestors.split('\n'), feature_ancestors.split('\n'))

    # The first line to match this is the ancestor we are looking for.
    pattern = re.compile(' [0-9a-f]{40}$')
    for line in ancestor_diff:
        if pattern.match(line):
            ancestor = line.strip()
            break

    # Just to be sure, explode as soon as possible.
    assert ancestor
    return ancestor



#!/usr/bin/env python
#
# This file is part of `gitflow`.
# Copyright (c) 2010-2011 Vincent Driessen
# Copyright (c) 2012 Hartmut Goebel
# Distributed under a BSD-like license. For full terms see the file LICENSE.txt
#

import os
import codecs
try:
    from setuptools import setup, find_packages
except ImportError:
    from ez_setup import use_setuptools
    use_setuptools()
    from setuptools import setup, find_packages

import gitflow as distmeta

long_description = codecs.open('README.rst', "r", "utf-8").read()

install_requires = [
    'GitPython>=0.3.2c1',
    'argparse',
    'setuptools',
    'xmlbuilder',
    'python-dateutil',
    'poster',
    'simplejson',
    'httplib2',
    'colorama',
    'rbtools',
    'gitdb>=0.5.1'
]

setup(
    name="salsita-gitflow",
    scripts=['bin/git-flow'],
    version=distmeta.__version__,
    description="Git extensions to provide high-level repository operations for Vincent Driessen's branching model.",
    author=distmeta.__author__,
    author_email=distmeta.__contact__,
    url=distmeta.__homepage__,
    platforms=["any"],
    license="BSD",
    packages=find_packages(),
    package_data={'':['gitflow/busyflow/pivotal/tzmap.txt']},
    include_package_data=True,
    install_requires=install_requires,
    zip_safe=False,
    classifiers=[
        # Picked from
        #    http://pypi.python.org/pypi?:action=list_classifiers
        #"Development Status :: 1 - Planning",
        #"Development Status :: 2 - Pre-Alpha",
        #"Development Status :: 3 - Alpha",
        "Development Status :: 4 - Beta",
        #"Development Status :: 5 - Production/Stable",
        #"Development Status :: 6 - Mature",
        #"Development Status :: 7 - Inactive",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Operating System :: POSIX",
        "Programming Language :: Python",
        "Topic :: Software Development :: Version Control",
    ],
    long_description=long_description,
)
